from __future__ import annotations

# Using pyright's assert type instead
# mypy: disable-error-code="assert-type"
from typing import TYPE_CHECKING, Any, cast

import pytest

import narwhals as nw

if TYPE_CHECKING:
    from narwhals._typing import (
        _ArrowImpl,
        _DaskImpl,
        _DuckDBImpl,
        _EagerAllowedImpl,
        _IbisImpl,
        _LazyAllowedImpl,
        _ModinImpl,
        _PandasImpl,
        _PolarsImpl,
        _SQLFrameImpl,
    )
    from narwhals.typing import IntoDataFrame, IntoLazyFrame, IntoSeries


def test_implementation_pandas() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    assert (
        nw.from_native(pd.DataFrame({"a": [1, 2, 3]})).implementation
        is nw.Implementation.PANDAS
    )
    assert (
        nw.from_native(pd.DataFrame({"a": [1, 2, 3]}))["a"].implementation
        is nw.Implementation.PANDAS
    )
    assert nw.from_native(pd.DataFrame({"a": [1, 2, 3]})).implementation.is_pandas()
    assert nw.from_native(pd.DataFrame({"a": [1, 2, 3]})).implementation.is_pandas_like()


def test_implementation_polars() -> None:
    pytest.importorskip("polars")
    import polars as pl

    assert not nw.from_native(pl.DataFrame({"a": [1, 2, 3]})).implementation.is_pandas()
    assert not nw.from_native(pl.DataFrame({"a": [1, 2, 3]}))[
        "a"
    ].implementation.is_pandas()
    assert nw.from_native(pl.DataFrame({"a": [1, 2, 3]})).implementation.is_polars()
    assert nw.from_native(pl.LazyFrame({"a": [1, 2, 3]})).implementation.is_polars()


@pytest.mark.parametrize(
    ("member", "value"),
    [
        ("PANDAS", "pandas"),
        ("MODIN", "modin"),
        ("CUDF", "cudf"),
        ("PYARROW", "pyarrow"),
        ("PYSPARK", "pyspark"),
        ("POLARS", "polars"),
        ("DASK", "dask"),
        ("DUCKDB", "duckdb"),
        ("IBIS", "ibis"),
        ("SQLFRAME", "sqlframe"),
        ("PYSPARK_CONNECT", "pyspark[connect]"),
        ("UNKNOWN", "unknown"),
    ],
)
def test_implementation_new(member: str, value: str) -> None:
    assert nw.Implementation(value) is getattr(nw.Implementation, member)


_TYPING_ONLY_TESTS = "_"
"""Exhaustive checks for overload matching native -> implementation.

## Arrange
Each test defines a function accepting a `native` frame.

Important:
    We *must* use the concrete types and therefore the type checker *needs* the package installed.

Next, wrap `native` as a `nw.{Data,Lazy}Frame`.

Note:
    If we support multiple native types, use `native` to generate `nw.{LazyFrame,Series}` as well.

Finally, look-up `.implementation` on all wrapped objects.

## Act
Try passing every result (`*_impl`) to functions that *only* accept a **subset** of `Implementation`.

This step *may require* a `# (type|pyright): ignore` directive, which defines the `# [... Negative]` result.
Otherwise, results are labelled with `# [... Positive]`.

If this *static* label matches *runtime* we use `# [True ...]`, otherwise `# [False ...]`.

Tip:
    `# [False Negative]`s are the most frustrating for users.
    Always try to minimize warning on safe code.

## Assert
The action determined whether or not our typing warns on an `@overload` match.

We still need to use [`assert_type`] to verify which `Implementation`(s) were returned as a result.

[`assert_type`]: https://typing-extensions.readthedocs.io/en/latest/#typing_extensions.assert_type
"""
if TYPE_CHECKING:
    import dask.dataframe as dd
    import duckdb
    import ibis
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from sqlframe.base.dataframe import BaseDataFrame
    from typing_extensions import assert_type

    import tests._modin_stub as mpd

    any_df: nw.DataFrame[Any] = cast("nw.DataFrame[Any]", "")
    any_ldf: nw.LazyFrame[Any] = cast("nw.LazyFrame[Any]", "")
    any_ser: nw.Series[Any] = cast("nw.Series[Any]", "")
    bound_df: nw.DataFrame[IntoDataFrame] = cast("nw.DataFrame[IntoDataFrame]", "")
    bound_ldf: nw.LazyFrame[IntoLazyFrame] = cast("nw.LazyFrame[IntoLazyFrame]", "")
    bound_ser: nw.Series[IntoSeries] = cast("nw.Series[IntoSeries]", "")

    def test_polars_typing(native: pl.DataFrame) -> None:
        df = nw.from_native(native)
        ldf = nw.from_native(native.lazy())
        ser = nw.from_native(native.to_series(), series_only=True)

        df_impl = df.implementation
        ldf_impl = ldf.implementation
        ser_impl = ser.implementation

        # [True Positive]
        any_df.lazy(df_impl)
        any_df.lazy(ldf_impl)
        any_df.lazy(ser_impl)
        any_ldf.collect(df_impl)
        any_ldf.collect(ldf_impl)
        any_ldf.collect(ser_impl)

        assert_type(df_impl, _PolarsImpl)
        assert_type(ldf_impl, _PolarsImpl)
        assert_type(ser_impl, _PolarsImpl)

    def test_pandas_typing(native: pd.DataFrame) -> None:
        df = nw.from_native(native)
        ldf = nw.from_native(native).lazy()
        ser = nw.from_native(native.iloc[0], series_only=True)

        df_impl = df.implementation
        ldf_impl = ldf.implementation
        ser_impl = ser.implementation

        # [True Negative]
        any_df.lazy(df_impl)  # type: ignore[arg-type]
        # [False Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_df.lazy(ser_impl)  # pyright: ignore[reportArgumentType]
        # [True Positive]
        any_ldf.collect(df_impl)
        any_ldf.collect(ldf_impl)
        any_ldf.collect(ser_impl)

        assert_type(df_impl, _PandasImpl)
        # NOTE: Would require adding overloads to `DataFrame.lazy`
        assert_type(ldf_impl, _PandasImpl)  # pyright: ignore[reportAssertTypeFailure]
        assert_type(ser_impl, _PandasImpl)

    def test_arrow_typing(native: pa.Table) -> None:
        df = nw.from_native(native)
        ldf = nw.from_native(native).lazy()
        ser = nw.from_native(native.column(0), series_only=True)

        df_impl = df.implementation
        ldf_impl = ldf.implementation
        ser_impl = ser.implementation

        # [True Negative]
        any_df.lazy(df_impl)  # type: ignore[arg-type]
        # [False Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_df.lazy(ser_impl)  # pyright: ignore[reportArgumentType]
        # [True Positive]
        any_ldf.collect(df_impl)
        any_ldf.collect(ldf_impl)
        any_ldf.collect(ser_impl)

        assert_type(df_impl, _ArrowImpl)
        # NOTE: Would require adding overloads to `DataFrame.lazy`
        assert_type(ldf_impl, _ArrowImpl)  # pyright: ignore[reportAssertTypeFailure]
        assert_type(ser_impl, _ArrowImpl)

    def test_duckdb_typing(native: duckdb.DuckDBPyRelation) -> None:
        ldf = nw.from_native(native)

        ldf_impl = ldf.implementation

        # [True Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_ldf.collect(ldf_impl)  # type: ignore[arg-type]

        assert_type(ldf.implementation, _DuckDBImpl)

    def test_sqlframe_typing(native: BaseDataFrame[Any, Any, Any, Any, Any]) -> None:
        ldf = nw.from_native(native)

        ldf_impl = ldf.implementation

        # [True Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_ldf.collect(ldf_impl)  # pyright: ignore[reportArgumentType]

        assert_type(ldf.implementation, _SQLFrameImpl)

    def test_ibis_typing(native: ibis.Table) -> None:
        ldf = nw.from_native(native)

        ldf_impl = ldf.implementation

        # [True Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_ldf.collect(ldf_impl)  # pyright: ignore[reportArgumentType]

        assert_type(ldf.implementation, _IbisImpl)

    def test_dask_typing(native: dd.DataFrame) -> None:
        ldf = nw.from_native(native)

        ldf_impl = ldf.implementation

        # [True Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_ldf.collect(ldf_impl)  # pyright: ignore[reportArgumentType]

        assert_type(ldf.implementation, _DaskImpl)

    def test_modin_typing(native: mpd.DataFrame) -> None:
        df = nw.from_native(native)
        # NOTE: Arbitrary method that returns a `Series`
        ser = nw.from_native(native.duplicated(), series_only=True)

        df_impl = df.implementation
        ser_impl = ser.implementation

        # [True Negative]
        any_df.lazy(df_impl)  # pyright: ignore[reportArgumentType]
        any_df.lazy(ser_impl)  # pyright: ignore[reportArgumentType]
        any_ldf.collect(df_impl)  # pyright: ignore[reportArgumentType]
        any_ldf.collect(ser_impl)  # pyright: ignore[reportArgumentType]

        assert_type(df_impl, _ModinImpl)
        assert_type(ser_impl, _ModinImpl)

    def test_any_typing() -> None:
        df_impl = any_df.implementation
        ldf_impl = any_ldf.implementation
        ser_impl = any_ser.implementation

        # [False Positive]
        any_df.lazy(df_impl)
        any_df.lazy(ldf_impl)
        any_df.lazy(ser_impl)
        any_ldf.collect(df_impl)
        any_ldf.collect(ldf_impl)
        any_ldf.collect(ser_impl)

        assert_type(df_impl, _EagerAllowedImpl)  # pyright: ignore[reportAssertTypeFailure]
        assert_type(ldf_impl, _LazyAllowedImpl)  # pyright: ignore[reportAssertTypeFailure]
        assert_type(ser_impl, _EagerAllowedImpl)  # pyright: ignore[reportAssertTypeFailure]
        # Fallback, matches the first overload `_PolarsImpl`
        assert_type(df_impl, _PolarsImpl)
        assert_type(ldf_impl, _PolarsImpl)
        assert_type(ser_impl, _PolarsImpl)

    def test_bound_typing() -> None:
        df_impl = bound_df.implementation
        ldf_impl = bound_ldf.implementation
        ser_impl = bound_ser.implementation

        # [True Negative]
        any_df.lazy(df_impl)  # type: ignore[arg-type]
        # [True Positive]
        any_df.lazy(ldf_impl)
        # [True Negative]
        any_df.lazy(ser_impl)  # type: ignore[arg-type]
        any_ldf.collect(df_impl)  # type: ignore[arg-type]
        any_ldf.collect(ldf_impl)  # type: ignore[arg-type]
        any_ldf.collect(ser_impl)  # type: ignore[arg-type]

        assert_type(df_impl, _EagerAllowedImpl)
        assert_type(ldf_impl, _LazyAllowedImpl)
        assert_type(ser_impl, _EagerAllowedImpl)

    def test_mixed_eager_typing(
        *args: nw.DataFrame[pl.DataFrame | pd.DataFrame | pa.Table]
        | nw.Series[pl.Series | pd.Series[Any] | pa.ChunkedArray[Any]],
    ) -> None:
        # NOTE: Any combination of eager objects that **does not** include `cuDF`, `modin` should
        # preserve that detail
        mix_impl = args[0].implementation

        # [True Negative]
        any_df.lazy(mix_impl)  # type: ignore[arg-type]
        # [True Positive]
        any_ldf.collect(mix_impl)

        assert_type(mix_impl, _PolarsImpl | _PandasImpl | _ArrowImpl)



"""Run ruff on Python examples in docstrings."""

from __future__ import annotations

import ast
import doctest
import os
import subprocess
import sys
import sysconfig
import tempfile
from pathlib import Path
from subprocess import CompletedProcess

SELECT = (
    "F",  # pyflakes-f
)
IGNORE = (
    "F811",  # redefined-while-unused
    "F821",  # undefined-name (misses https://docs.python.org/3/library/doctest.html#what-s-the-execution-context)
)


def find_ruff_bin() -> Path:
    """Return the ruff binary path.

    Adapted from [`ruff.__main__.find_ruff_bin`], see also [astral-sh/ruff#18153], [astral-sh/uv#1677].

    [`ruff.__main__.find_ruff_bin`]: https://github.com/astral-sh/ruff/blob/2d6ca092fa1655f14f10dab6e2a5b95f5f682c24/python/ruff/__main__.py
    [astral-sh/ruff#18153]: https://github.com/astral-sh/ruff/issues/18153#issuecomment-2888581114
    [astral-sh/uv#1677]: https://github.com/astral-sh/uv/issues/1677
    """
    ruff_exe: str = "ruff" + sysconfig.get_config_var("EXE")

    scripts_path = Path(sysconfig.get_path("scripts")) / ruff_exe
    if scripts_path.is_file():
        return scripts_path

    if sys.version_info >= (3, 10):
        user_scheme = sysconfig.get_preferred_scheme("user")
    elif os.name == "nt":
        user_scheme = "nt_user"
    elif sys.platform == "darwin" and sys._framework:
        user_scheme = "osx_framework_user"
    else:
        user_scheme = "posix_user"

    user_path = Path(sysconfig.get_path("scripts", scheme=user_scheme)) / ruff_exe
    if user_path.is_file():
        return user_path
    msg = (
        f"Unable to find ruff at:\n- {scripts_path.as_posix()}\n- {user_path.as_posix()}\n\n"
        "Hint: did you follow this guide? https://github.com/narwhals-dev/narwhals?tab=contributing-ov-file#readme"
    )
    raise FileNotFoundError(msg)


def extract_docstring_examples(files: list[str]) -> list[tuple[Path, str, str]]:
    """Extract examples from docstrings in Python files."""
    examples: list[tuple[Path, str, str]] = []

    for file in files:  # noqa: PLR1702
        fp = Path(file)
        tree = ast.parse(fp.read_text("utf-8"))

        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
                docstring = ast.get_docstring(node)
                if docstring:
                    parsed_examples = doctest.DocTestParser().get_examples(docstring)
                    example_code = "\n".join(
                        example.source for example in parsed_examples
                    )
                    if example_code.strip():
                        examples.append((fp, node.name, example_code))

    return examples


def create_temp_files(examples: list[tuple[Path, str, str]]) -> list[tuple[Path, str]]:
    """Create temporary files for all examples and return their paths."""
    temp_files: list[tuple[Path, str]] = []

    for file, name, example in examples:
        temp_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            encoding="utf-8", mode="w", suffix=".py", delete=False
        )
        temp_file.write(example)
        temp_file_path = temp_file.name
        temp_file.close()
        temp_files.append((Path(temp_file_path), f"{file!s}:{name}"))

    return temp_files


def run_ruff_on_temp_files(
    temp_files: list[tuple[Path, str]],
) -> CompletedProcess[str] | None:
    """Run ruff on all temporary files and collect error messages."""
    temp_file_paths = [temp_file[0] for temp_file in temp_files]
    select = f"--select={','.join(SELECT)}"
    ignore = f"--ignore={','.join(IGNORE)}"
    result = subprocess.run(  # noqa: S603
        [find_ruff_bin(), "check", select, ignore, *temp_file_paths],
        capture_output=True,
        text=True,
        check=False,
    )

    if result.returncode == 0:
        return None
    return result


def report_errors(
    completed: CompletedProcess[str] | None, temp_files: list[tuple[Path, str]]
) -> None:
    """Map errors back to original examples and report them."""
    if completed is None:
        return

    print("Ruff issues found in examples:\n")
    stdout = completed.stdout
    for temp_file, original_context in temp_files:
        stdout = stdout.replace(str(temp_file), original_context)
    print(stdout)


def cleanup_temp_files(temp_files: list[tuple[Path, str]]) -> None:
    """Remove all temporary files."""
    for temp_file, _ in temp_files:
        temp_file.unlink()


def main(python_files: list[str]) -> None:
    docstring_examples = extract_docstring_examples(python_files)

    if not docstring_examples:
        sys.exit(0)

    temp_files = create_temp_files(docstring_examples)

    try:
        errors = run_ruff_on_temp_files(temp_files)
        report_errors(errors, temp_files)
    finally:
        cleanup_temp_files(temp_files)

    if errors:
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(line_item_ds: FrameT) -> FrameT:
    var_1 = datetime(1994, 1, 1)
    var_2 = datetime(1995, 1, 1)
    var_3 = 24

    return (
        line_item_ds.filter(
            nw.col("l_shipdate").is_between(var_1, var_2, closed="left"),
            nw.col("l_discount").is_between(0.05, 0.07),
            nw.col("l_quantity") < var_3,
        )
        .with_columns((nw.col("l_extendedprice") * nw.col("l_discount")).alias("revenue"))
        .select(nw.sum("revenue"))
    )



from __future__ import annotations

from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from narwhals.exceptions import InvalidOperationError
from tests.utils import (
    DUCKDB_VERSION,
    POLARS_VERSION,
    PYARROW_VERSION,
    Constructor,
    assert_equal_data,
)

if TYPE_CHECKING:
    from narwhals.typing import PythonLiteral
    from tests.utils import ConstructorEager

data: dict[str, list[PythonLiteral]] = {
    "a": [8, 2, 1, None],
    "b": [58, 5, 6, 12],
    "c": [2.5, 1.0, 3.0, 0.9],
    "d": [2, 1, 4, 3],
    "idx": [0, 1, 2, 3],
}

single_cases = pytest.mark.parametrize(
    ("col", "expected_first", "expected_last"),
    [("a", 8, None), ("b", 58, 12), ("c", 2.5, 0.9)],
)


@single_cases
def test_first_last_series(
    constructor_eager: ConstructorEager,
    col: str,
    expected_first: PythonLiteral,
    expected_last: PythonLiteral,
) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)[col]
    result = series.first()
    assert_equal_data({col: [result]}, {col: [expected_first]})
    result = series.last()
    assert_equal_data({col: [result]}, {col: [expected_last]})


def test_first_last_series_empty(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)["a"]
    series = series.filter(series > 50)
    result = series.first()
    assert result is None
    result = series.last()
    assert result is None


@single_cases
def test_first_last_expr_select(
    constructor_eager: ConstructorEager,
    col: str,
    expected_first: PythonLiteral,
    expected_last: PythonLiteral,
) -> None:
    df = nw.from_native(constructor_eager(data))
    result = df.select(
        nw.col(col).first().name.suffix("_first"), nw.col(col).last().name.suffix("_last")
    )
    assert_equal_data(
        result, {f"{col}_first": [expected_first], f"{col}_last": [expected_last]}
    )


@single_cases
def test_first_expr_with_columns(
    constructor_eager: ConstructorEager,
    col: str,
    expected_first: PythonLiteral,
    expected_last: PythonLiteral,
) -> None:
    df = nw.from_native(constructor_eager(data))
    result = df.with_columns(
        nw.col(col).first().name.suffix("_first"), nw.col(col).last().name.suffix("_last")
    ).select(nw.selectors.matches("first|last"))
    assert_equal_data(
        result,
        {
            f"{col}_first": [expected_first] * len(data["a"]),
            f"{col}_last": [expected_last] * len(data["a"]),
        },
    )


def test_first_last_expr_over_order_by(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 10):
        pytest.skip()
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()

    if any(x in str(constructor) for x in ("pyspark", "dask")):
        # Currently unsupported.
        request.applymarker(pytest.mark.xfail)
    if "cudf" in str(constructor):
        reason = "Need to pass dtype when passing pd.NA or None"
        request.applymarker(pytest.mark.xfail(reason=reason))

    frame = nw.from_native(
        constructor(
            {
                "a": [1, 1, 2],
                "b": [4, 5, 6],
                "c": [None, 7, 8],
                "d": ["x", "y", "z"],
                "i": [None, 2, 1],
            }
        )
    )
    result = frame.with_columns(
        nw.col("b", "c", "d")
        .first()
        .over(order_by="i")
        .name.suffix("_first_ordered_over"),
        nw.col("b", "c", "d").last().over(order_by="i").name.suffix("_last_ordered_over"),
        nw.col("b", "c", "d").first(order_by="i").name.suffix("_first"),
        nw.col("b", "c", "d").last(order_by="i").name.suffix("_last"),
    ).sort("i")
    expected = {
        "a": [1, 2, 1],
        "b": [4, 6, 5],
        "c": [None, 8, 7],
        "d": ["x", "z", "y"],
        "i": [None, 1, 2],
        "b_first_ordered_over": [4, 4, 4],
        "c_first_ordered_over": [None, None, None],
        "d_first_ordered_over": ["x", "x", "x"],
        "b_last_ordered_over": [5, 5, 5],
        "c_last_ordered_over": [7, 7, 7],
        "d_last_ordered_over": ["y", "y", "y"],
        "b_first": [4, 4, 4],
        "c_first": [None, None, None],
        "d_first": ["x", "x", "x"],
        "b_last": [5, 5, 5],
        "c_last": [7, 7, 7],
        "d_last": ["y", "y", "y"],
    }
    assert_equal_data(result, expected)


def test_first_expr_over_order_by_partition_by(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 10):
        pytest.skip()
    if "pyarrow_table" in str(constructor) and PYARROW_VERSION < (14,):
        pytest.skip()
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()

    if any(x in str(constructor) for x in ("pyspark", "dask")):
        # Currently unsupported.
        request.applymarker(pytest.mark.xfail)
    if "ibis" in str(constructor):
        # https://github.com/ibis-project/ibis/issues/11656
        request.applymarker(pytest.mark.xfail)

    frame = nw.from_native(
        constructor(
            {"a": [1, 1, 2], "b": [4, 5, 6], "c": [None, 7, 8], "i": [1, None, 2]}
        )
    )
    result = frame.with_columns(
        nw.col("b", "c").first().over("a", order_by="i").name.suffix("_first"),
        nw.col("b", "c").last().over("a", order_by="i").name.suffix("_last"),
    ).sort("i")
    expected = {
        "a": [1, 1, 2],
        "b": [5, 4, 6],
        "c": [7.0, None, 8.0],
        "i": [None, 1, 2],
        "b_first": [5, 5, 6],
        "c_first": [7.0, 7.0, 8.0],
        "b_last": [4, 4, 6],
        "c_last": [None, None, 8.0],
    }
    assert_equal_data(result, expected)


def test_first_expr_in_group_by(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("spark", "dask")):
        # ibis: https://github.com/ibis-project/ibis/issues/11656
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    if "pyarrow_table" in str(constructor) and PYARROW_VERSION < (14,):
        pytest.skip()
    data = {
        "grp": [1, 1, 1, 2],
        "a": [None, 4, 9, 3],
        "b": [9, 7, 10, 8],
        "c": [9, None, 10, 8],
        "idx": [9, None, None, 7],
        "idx2": [9, 8, 7, 7],
    }
    df = nw.from_native(constructor(data))
    result = (
        df.group_by("grp")
        .agg(
            nw.col("a", "b", "c").first(order_by="idx").name.suffix("_first"),
            nw.col("a", "b", "c").last(order_by="idx").name.suffix("_last"),
        )
        .sort("grp")
    )
    expected = {
        "grp": [1, 2],
        "a_first": [4, 3],
        "b_first": [7, 8],
        "c_first": [None, 8],
        "a_last": [None, 3],
        "b_last": [9, 8],
        "c_last": [9, 8],
    }
    assert_equal_data(result, expected)
    result = (
        df.group_by("grp")
        .agg(
            nw.col("a", "b", "c").first(order_by=["idx", "idx2"]).name.suffix("_first"),
            nw.col("a", "b", "c").last(order_by=["idx", "idx2"]).name.suffix("_last"),
        )
        .sort("grp")
    )
    expected = {
        "grp": [1, 2],
        "a_first": [9, 3],
        "b_first": [10, 8],
        "c_first": [10, 8],
        "a_last": [None, 3],
        "b_last": [9, 8],
        "c_last": [9, 8],
    }
    assert_equal_data(result, expected)


def test_first_expr_broadcasting(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("spark", "dask")):
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()

    data = {
        "grp": [1, 1, 1, 2],
        "a": [None, 4, 9, 3],
        "b": [9, 7, 10, 8],
        "c": [9, None, 10, 8],
        "idx": [9, None, None, 7],
        "idx2": [9, 8, 7, 7],
    }
    df = nw.from_native(constructor(data))
    result = df.select(
        "idx", "idx2", nw.col("a", "b", "c").first(order_by="idx").name.suffix("_first")
    ).sort("idx", "idx2")
    expected = {
        "idx": [None, None, 7, 9],
        "idx2": [7, 8, 7, 9],
        "a_first": [4, 4, 4, 4],
        "b_first": [7, 7, 7, 7],
        "c_first": [None, None, None, None],
    }
    assert_equal_data(result, expected)
    result = df.select(
        "idx",
        "idx2",
        nw.col("a", "b", "c").first(order_by=["idx", "idx2"]).name.suffix("_first"),
    ).sort("idx", "idx2")
    expected = {
        "idx": [None, None, 7, 9],
        "idx2": [7, 8, 7, 9],
        "a_first": [9, 9, 9, 9],
        "b_first": [10, 10, 10, 10],
        "c_first": [10, 10, 10, 10],
    }
    assert_equal_data(result, expected)


def test_first_expr_invalid(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "dask" in str(constructor):
        request.applymarker(pytest.mark.xfail)
    data = {
        "grp": [1, 1, 1, 2],
        "a": [None, 4, 9, 3],
        "b": [9, 7, 10, 8],
        "c": [9, None, 10, 8],
        "idx": [9, None, None, 7],
        "idx2": [9, 8, 7, 7],
    }
    df = nw.from_native(constructor(data))
    with pytest.raises(InvalidOperationError, match="expression which isn't orderable"):
        df.select("idx", "idx2", nw.col("a").first(order_by="idx").over(order_by="idx2"))


def test_first_last_different_orders(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 10):
        pytest.skip()
    if any(x in str(constructor) for x in ("pyspark", "dask")):
        # Currently unsupported.
        request.applymarker(pytest.mark.xfail)
    if "pyarrow_table" in str(constructor) and PYARROW_VERSION < (14,):
        pytest.skip()
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()

    context = (
        pytest.raises(NotImplementedError, match="Only one `order_by` can")
        if "pandas" in str(constructor) or "pyarrow" in str(constructor)
        else does_not_raise()
    )
    frame = nw.from_native(
        constructor(
            {
                "a": [1, 1, 2],
                "b": [4, 5, 6],
                "c": [None, 7, 8],
                "i_0": [1, None, 2],
                "i_1": [2, None, 1],
            }
        )
    )
    with context:
        result = (
            frame.group_by("a")
            .agg(
                nw.col("b", "c").first(order_by="i_0").name.suffix("_first_i_0"),
                nw.col("b", "c").first(order_by="i_1").name.suffix("_first_i_1"),
                nw.col("b", "c").last(order_by="i_0").name.suffix("_last_i_0"),
                nw.col("b", "c").last(order_by="i_1").name.suffix("_last_i_1"),
            )
            .sort("a")
        )
        expected = {
            "a": [1, 2],
            "b_first_i_0": [5, 6],
            "c_first_i_0": [7, 8],
            "b_first_i_1": [5, 6],
            "c_first_i_1": [7, 8],
            "b_last_i_0": [4, 6],
            "c_last_i_0": [None, 8],
            "b_last_i_1": [4, 6],
            "c_last_i_1": [None, 8],
        }
        assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from narwhals.dependencies import is_narwhals_series

if TYPE_CHECKING:
    from typing_extensions import Never, TypeAlias

    # NOTE: These aliases are created to facilitate autocompletion.
    # Feel free to extend them as you please when adding new features.
    # See: https://github.com/narwhals-dev/narwhals/pull/2983#discussion_r2337548736
    ObjectName: TypeAlias = Literal["inputs", "Series", "DataFrames"]
    SeriesDetail: TypeAlias = Literal[
        "dtype mismatch",
        "exact value mismatch",
        "implementation mismatch",
        "length mismatch",
        "name mismatch",
        "nested value mismatch",
        "null value mismatch",
        "values not within tolerance",
    ]
    DataFramesDetail: TypeAlias = Literal[
        "columns are not in the same order",
        "dtypes do not match",
        "height (row count) mismatch",
        "implementation mismatch",
        "in left, but not in right",
        "in right, but not in left",
        "value mismatch for column",
    ]


def raise_assertion_error(
    objects: ObjectName,
    detail: str,
    left: Any,
    right: Any,
    *,
    cause: Exception | None = None,
) -> Never:
    """Raise a detailed assertion error."""
    __tracebackhide__ = True

    trailing_left = "\n" if is_narwhals_series(left) else " "
    trailing_right = "\n" if is_narwhals_series(right) else " "

    msg = (
        f"{objects} are different ({detail})\n"
        f"[left]:{trailing_left}{left}\n"
        f"[right]:{trailing_right}{right}"
    )
    raise AssertionError(msg) from cause


def raise_series_assertion_error(
    detail: SeriesDetail, left: Any, right: Any, *, cause: Exception | None = None
) -> Never:
    raise_assertion_error("Series", detail, left, right, cause=cause)


def raise_frame_assertion_error(
    detail: DataFramesDetail,
    left: Any,
    right: Any,
    *,
    detail_prefix: str = "",
    detail_suffix: str = "",
    cause: Exception | None = None,
) -> Never:
    raise_assertion_error(
        "DataFrames", f"{detail_prefix}{detail}{detail_suffix}", left, right, cause=cause
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from narwhals._compliant import CompliantNamespace
from narwhals._utils import not_implemented
from test_plugin.dataframe import DictFrame, DictLazyFrame

if TYPE_CHECKING:
    from narwhals.utils import Version


class DictNamespace(CompliantNamespace[DictLazyFrame, Any]):
    def __init__(self, *, version: Version) -> None:
        self._version = version

    def from_native(self, native_object: DictFrame) -> DictLazyFrame:
        return DictLazyFrame(native_object, version=self._version)

    is_native: Any = not_implemented()
    _expr: Any = not_implemented()
    _implementation: Any = not_implemented()
    corr: Any = not_implemented()
    len: Any = not_implemented()
    lit: Any = not_implemented()
    all_horizontal: Any = not_implemented()
    any_horizontal: Any = not_implemented()
    sum_horizontal: Any = not_implemented()
    mean_horizontal: Any = not_implemented()
    min_horizontal: Any = not_implemented()
    max_horizontal: Any = not_implemented()
    concat: Any = not_implemented()
    when: Any = not_implemented()
    concat_str: Any = not_implemented()
    selectors: Any = not_implemented()
    coalesce: Any = not_implemented()
    struct: Any = not_implemented()



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import DASK_VERSION, Constructor, assert_equal_data


def test_expr_binary(constructor: Constructor) -> None:
    if "dask" in str(constructor) and DASK_VERSION < (2024, 10):
        pytest.skip()
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_raw = constructor(data)
    result = nw.from_native(df_raw).with_columns(
        a=(1 + 3 * nw.col("a")) * (1 / nw.col("a")),
        b=nw.col("z") / (2 - nw.col("b")),
        c=nw.col("a") + nw.col("b") / 2,
        d=nw.col("a") - nw.col("b"),
        e=((nw.col("a") > nw.col("b")) & (nw.col("a") >= nw.col("z"))).cast(nw.Int64),
        f=(
            (nw.col("a") < nw.col("b"))
            | (nw.col("a") <= nw.col("z"))
            | (nw.col("a") == 1)
        ).cast(nw.Int64),
        g=nw.col("a") != 1,
        h=(False & (nw.col("a") != 1)),
        i=(False | (nw.col("a") != 1)),
        j=2 ** nw.col("a"),
        k=2 // nw.col("a"),
        l=nw.col("a") // 2,
        m=nw.col("a") ** 2,
    )
    expected = {
        "a": [4, 3.333333, 3.5],
        "b": [-3.5, -4.0, -2.25],
        "z": [7.0, 8.0, 9.0],
        "c": [3, 5, 5],
        "d": [-3, -1, -4],
        "e": [0, 0, 0],
        "f": [1, 1, 1],
        "g": [False, True, True],
        "h": [False, False, False],
        "i": [False, True, True],
        "j": [2, 8, 4],
        "k": [2, 0, 1],
        "l": [0, 1, 1],
        "m": [1, 9, 4],
    }
    assert_equal_data(result, expected)



from __future__ import annotations

from typing import Any

import pytest

import narwhals as nw
from tests.utils import DUCKDB_VERSION, Constructor, assert_equal_data


def test_sumh(constructor: Constructor) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df = nw.from_native(constructor(data))
    result = df.with_columns(horizontal_sum=nw.sum_horizontal("a", nw.col("b")))
    expected = {
        "a": [1, 3, 2],
        "b": [4, 4, 6],
        "z": [7.0, 8.0, 9.0],
        "horizontal_sum": [5, 7, 8],
    }
    assert_equal_data(result, expected)


def test_sumh_nullable(constructor: Constructor) -> None:
    data = {"a": [1, 8, 3], "b": [4, 5, None], "i": [0, 1, 2]}
    expected = {"hsum": [5, 13, 3], "i": [0, 1, 2]}

    df = nw.from_native(constructor(data))
    result = df.select(hsum=nw.sum_horizontal("a", "b"), i="i").sort("i")
    assert_equal_data(result, expected)


def test_sumh_all(constructor: Constructor) -> None:
    data = {"a": [1, 2, 3], "b": [10, 20, 30]}
    df = nw.from_native(constructor(data))
    result = df.select(nw.sum_horizontal(nw.all().name.suffix("_foo")))
    expected = {"a_foo": [11, 22, 33]}
    assert_equal_data(result, expected)
    result = df.select(c=nw.sum_horizontal(nw.all()))
    expected = {"c": [11, 22, 33]}
    assert_equal_data(result, expected)


def test_sumh_aggregations(constructor: Constructor) -> None:
    data = {"a": [1, 2, 3], "b": [10, 20, 30]}
    df = nw.from_native(constructor(data))
    result = df.select(nw.sum_horizontal(nw.all().mean().name.suffix("_foo")))
    expected = {"a_foo": [22]}
    assert_equal_data(result, expected)


def test_sumh_transformations(constructor: Constructor) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
    df = nw.from_native(constructor(data))
    result = df.select(d=nw.sum_horizontal("a", nw.col("b").sum(), "c"))
    expected: dict[str, Any] = {"d": [23, 25, 27]}
    assert_equal_data(result, expected)
    result = df.select(d=nw.sum_horizontal("a", nw.lit(None, dtype=nw.Float64), "c"))
    expected = {"d": [8.0, 10.0, 12.0]}
    assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from narwhals._utils import (
    Implementation,
    ValidateBackendVersion,
    Version,
    not_implemented,
)
from narwhals.typing import CompliantLazyFrame

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias

    from narwhals import LazyFrame  # noqa: F401

DictFrame: TypeAlias = dict[str, list[Any]]


class DictLazyFrame(
    CompliantLazyFrame[Any, "DictFrame", "LazyFrame[DictFrame]"],  # type: ignore[type-var]
    ValidateBackendVersion,
):
    _implementation = Implementation.UNKNOWN

    def __init__(self, native_dataframe: DictFrame, *, version: Version) -> None:
        self._native_frame: DictFrame = native_dataframe
        self._version = version

    def __narwhals_lazyframe__(self) -> Self:
        return self

    @property
    def columns(self) -> list[str]:  # pragma: no cover
        return list(self._native_frame.keys())

    _with_native = not_implemented()

    def _with_version(self, version: Version) -> Self:
        return self.__class__(self._native_frame, version=version)

    # Dunders
    __narwhals_namespace__ = not_implemented()
    __native_namespace__ = not_implemented()

    # Properties
    schema = not_implemented()  # type: ignore[assignment]

    # Static
    _is_native = not_implemented()  # type: ignore[assignment]

    # Helpers
    _iter_columns = not_implemented()

    # Functions
    aggregate = not_implemented()
    collect = not_implemented()
    collect_schema = not_implemented()
    drop = not_implemented()
    drop_nulls = not_implemented()
    explode = not_implemented()
    filter = not_implemented()
    from_native = not_implemented()
    group_by = not_implemented()
    head = not_implemented()
    join = not_implemented()
    join_asof = not_implemented()
    rename = not_implemented()
    select = not_implemented()
    simple_select = not_implemented()
    sink_parquet = not_implemented()
    sort = not_implemented()
    tail = not_implemented()
    to_narwhals = not_implemented()
    unique = not_implemented()
    unpivot = not_implemented()
    with_columns = not_implemented()
    with_row_index = not_implemented()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

pytest.importorskip("numpy")

import narwhals as nw
from tests.utils import (
    PANDAS_VERSION,
    POLARS_VERSION,
    ConstructorEager,
    assert_equal_data,
)

if TYPE_CHECKING:
    from narwhals.dtypes import DType

data = {"a": [1, 2, 3], "b": [4, 5, 6], "z": [7.0, 8.0, 9.0]}


def test_map_batches_expr_compliant(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    expected = df.select(nw.col("a", "b").map_batches(lambda s: s + 1).name.suffix("1"))
    assert_equal_data(expected, {"a1": [2, 3, 4], "b1": [5, 6, 7]})


@pytest.mark.parametrize(
    ("value", "dtype"),
    [(1, nw.Int64()), ("foo", nw.String()), ([1, 2], nw.List(nw.Int64()))],
)
def test_map_batches_expr_scalar(
    constructor_eager: ConstructorEager, value: Any, dtype: DType
) -> None:
    df = nw.from_native(constructor_eager(data))
    if dtype.is_nested() and df.implementation.is_pandas_like():
        if PANDAS_VERSION < (2, 2):  # pragma: no cover
            reason = "pandas is too old for nested types"
            pytest.skip(reason=reason)
        pytest.importorskip("pyarrow")

    expected = df.select(
        nw.col("a", "b").map_batches(
            lambda _: value, returns_scalar=True, return_dtype=dtype
        )
    )
    assert_equal_data(expected, {"a": [value], "b": [value]})


def test_map_batches_expr_numpy_array(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    expected = df.select(
        nw.col("a")
        .map_batches(lambda s: s.to_numpy() + 1, return_dtype=nw.Float64())
        .sum()
    )
    assert_equal_data(expected, {"a": [9.0]})


def test_map_batches_expr_numpy_scalar(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))

    expected = df.select(
        nw.all().map_batches(lambda s: s.to_numpy().argmax(), returns_scalar=True)
    )
    assert_equal_data(expected, {"a": [2], "b": [2], "z": [2]})


def test_map_batches_expr_names(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data))
    expected = nw.from_native(df.select(nw.all().map_batches(lambda x: x.to_numpy())))
    assert_equal_data(expected, {"a": [1, 2, 3], "b": [4, 5, 6], "z": [7.0, 8.0, 9.0]})


def test_map_batches_exception(
    constructor_eager: ConstructorEager, request: pytest.FixtureRequest
) -> None:
    if "polars" in str(constructor_eager) and POLARS_VERSION < (1, 32, 3):
        request.applymarker(pytest.mark.xfail)

    df = nw.from_native(constructor_eager(data))
    msg = (
        r"`map(?:_batches)?` with `returns_scalar=False` must return a Series; found "
        r"'numpy\.int(?:32|64)'\."
        "\n\nIf `returns_scalar` is set to `True`, a returned value can be "
        "a scalar value."
    )

    with pytest.raises(TypeError, match=msg):
        df.select(nw.all().map_batches(lambda s: s.to_numpy().argmax()))



from __future__ import annotations

import ast
import sys
from pathlib import Path

BANNED_IMPORTS = {
    "cudf",
    "cupy",
    "dask",
    "dask.dataframe",
    "dask_expr",
    "duckdb",
    "ibis",
    "modin",
    "numpy",
    "pandas",
    "polars",
    "pyarrow",
    "pyspark",
}

ALLOWED_IMPORTS = {
    "_pandas_like": {"pandas", "numpy"},
    "_arrow": {"pyarrow", "pyarrow.compute", "pyarrow.parquet"},
    "_dask": {"dask.dataframe", "pandas", "dask_expr"},
    "_polars": {"polars"},
    "_duckdb": {"duckdb"},
    "_ibis": {"ibis", "ibis._", "ibis.expr.types"},
}


class ImportPandasChecker(ast.NodeVisitor):
    def __init__(self, file_name: str, lines: list[str]) -> None:
        self.file_name = file_name
        self.lines = lines
        self.found_import = False
        for key, val in ALLOWED_IMPORTS.items():
            if key in self.file_name:
                self.allowed_imports: set[str] = val
                break
        else:
            self.allowed_imports = set()

    def visit_If(self, node: ast.If) -> None:
        # Check if the condition is `if TYPE_CHECKING`
        if isinstance(node.test, ast.Name) and node.test.id == "TYPE_CHECKING":
            # Skip the body of this if statement
            return
        self.generic_visit(node)

    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            if (
                alias.name in BANNED_IMPORTS
                and alias.name not in self.allowed_imports
                and "# ignore-banned-import" not in self.lines[node.lineno - 1]
            ):
                print(
                    f"{self.file_name}:{node.lineno}:{node.col_offset}: found {alias.name} import"
                )
                self.found_import = True

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if (
            node.module in BANNED_IMPORTS
            and "# ignore-banned-import" not in self.lines[node.lineno - 1]
            and node.module not in self.allowed_imports
        ):
            print(
                f"{self.file_name}:{node.lineno}:{node.col_offset}: found {node.module} import"
            )
            self.found_import = True
        self.generic_visit(node)


def check_import_pandas(filename: str) -> bool:
    content = Path(filename).read_text("utf-8")
    tree = ast.parse(content, filename=filename)

    checker = ImportPandasChecker(filename, content.splitlines())
    checker.visit(tree)

    return checker.found_import


if __name__ == "__main__":
    ret = 0
    for filename in sys.argv[1:]:
        if not filename.endswith(".py"):
            continue
        ret |= check_import_pandas(filename)
    sys.exit(ret)



from __future__ import annotations

import datetime as dt
import re
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals._utils import Implementation
from narwhals.exceptions import InvalidOperationError, NarwhalsError
from tests.utils import POLARS_VERSION, Constructor, ConstructorEager, assert_equal_data

if TYPE_CHECKING:
    from collections.abc import Iterator


def test_concat_horizontal(constructor_eager: ConstructorEager) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_left = nw.from_native(constructor_eager(data), eager_only=True)

    data_right = {"c": [6, 12, -1], "d": [0, -4, 2]}
    df_right = nw.from_native(constructor_eager(data_right), eager_only=True)

    result = nw.concat([df_left, df_right], how="horizontal")
    expected = {
        "a": [1, 3, 2],
        "b": [4, 4, 6],
        "z": [7.0, 8.0, 9.0],
        "c": [6, 12, -1],
        "d": [0, -4, 2],
    }
    assert_equal_data(result, expected)

    with pytest.raises(ValueError, match="No items"):
        nw.concat([])
    pattern = re.compile(r"horizontal.+not supported.+lazyframe", re.IGNORECASE)
    with pytest.raises(InvalidOperationError, match=pattern):
        nw.concat([df_left.lazy()], how="horizontal")


def test_concat_vertical(constructor: Constructor) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_left = (
        nw.from_native(constructor(data)).lazy().rename({"a": "c", "b": "d"}).drop("z")
    )

    data_right = {"c": [6, 12, -1], "d": [0, -4, 2]}
    df_right = nw.from_native(constructor(data_right)).lazy()

    result = nw.concat([df_left, df_right], how="vertical")
    expected = {"c": [1, 3, 2, 6, 12, -1], "d": [4, 4, 6, 0, -4, 2]}
    assert_equal_data(result, expected)

    with pytest.raises(ValueError, match="No items"):
        nw.concat([], how="vertical")

    with pytest.raises(
        (Exception, TypeError),
        match=r"unable to vstack|inputs should all have the same schema",
    ):
        nw.concat([df_left, df_right.rename({"d": "i"})], how="vertical").collect()
    with pytest.raises(
        (Exception, TypeError),
        match=r"unable to vstack|unable to append|inputs should all have the same schema",
    ):
        nw.concat([df_left, df_left.select("d")], how="vertical").collect()


def test_concat_diagonal(constructor: Constructor) -> None:
    data_1 = {"a": [1, 3], "b": [4, 6]}
    data_2 = {"a": [100, 200], "z": ["x", "y"]}
    expected = {
        "a": [1, 3, 100, 200],
        "b": [4, 6, None, None],
        "z": [None, None, "x", "y"],
    }

    df_1 = nw.from_native(constructor(data_1)).lazy()
    df_2 = nw.from_native(constructor(data_2)).lazy()

    result = nw.concat([df_1, df_2], how="diagonal")

    assert_equal_data(result, expected)

    with pytest.raises(ValueError, match="No items"):
        nw.concat([], how="diagonal")


def _from_natives(
    constructor: Constructor, *sources: dict[str, list[Any]]
) -> Iterator[nw.LazyFrame[Any]]:
    yield from (nw.from_native(constructor(data)).lazy() for data in sources)


def test_concat_diagonal_bigger(constructor: Constructor) -> None:
    # NOTE: `ibis.union` doesn't guarantee the order of outputs
    # https://github.com/narwhals-dev/narwhals/pull/3404#discussion_r2694556781
    data_1 = {"idx": [1, 2], "a": [1, 2], "b": [3, 4]}
    data_2 = {"a": [5, 6], "c": [7, 8], "idx": [3, 4]}
    data_3 = {"b": [9, 10], "idx": [5, 6], "c": [11, 12]}
    expected = {
        "idx": [1, 2, 3, 4, 5, 6],
        "a": [1, 2, 5, 6, None, None],
        "b": [3, 4, None, None, 9, 10],
        "c": [None, None, 7, 8, 11, 12],
    }
    dfs = _from_natives(constructor, data_1, data_2, data_3)
    result = nw.concat(dfs, how="diagonal").sort("idx")
    assert_equal_data(result, expected)


def test_concat_diagonal_invalid(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    data_1 = {"a": [1, 3], "b": [4, 6]}
    data_2 = {
        "a": [dt.datetime(2000, 1, 1), dt.datetime(2000, 1, 2)],
        "b": [4, 6],
        "z": ["x", "y"],
    }
    df_1 = nw.from_native(constructor(data_1)).lazy()
    bad_schema = nw.from_native(constructor(data_2)).lazy()
    impl = df_1.implementation
    request.applymarker(
        pytest.mark.xfail(
            impl not in {Implementation.IBIS, Implementation.POLARS},
            reason=f"{impl!r} does not validate schemas for `concat(how='diagonal')",
        )
    )
    context: Any
    if impl.is_polars() and POLARS_VERSION < (1, 18):  # pragma: no cover
        context = pytest.raises(
            NarwhalsError,
            match=re.compile(r"(int.+datetime)|(datetime.+int)", re.IGNORECASE),
        )
    else:
        context = pytest.raises((InvalidOperationError, TypeError), match=r"same schema")
    with context:
        nw.concat([df_1, bad_schema], how="diagonal").collect().to_dict(as_series=False)



from __future__ import annotations

import functools
import operator
import re
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, cast

import numpy as np
import pandas as pd

from narwhals._compliant import EagerSeriesNamespace
from narwhals._constants import (
    MS_PER_SECOND,
    NS_PER_MICROSECOND,
    NS_PER_MILLISECOND,
    NS_PER_SECOND,
    SECONDS_PER_DAY,
    US_PER_SECOND,
)
from narwhals._exceptions import issue_warning
from narwhals._utils import (
    Implementation,
    Version,
    _DeferredIterable,
    check_columns_exist,
    isinstance_or_issubclass,
    parse_version,
    requires,
)
from narwhals.exceptions import ShapeError

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping
    from types import ModuleType

    import pyarrow as pa
    from pandas._typing import Dtype as PandasDtype
    from pandas.core.dtypes.dtypes import BaseMaskedDtype
    from typing_extensions import TypeAlias, TypeIs

    from narwhals._duration import IntervalUnit
    from narwhals._pandas_like.expr import PandasLikeExpr
    from narwhals._pandas_like.series import PandasLikeSeries
    from narwhals._pandas_like.typing import (
        NativeDataFrameT,
        NativeNDFrameT,
        NativeSeriesT,
    )
    from narwhals.dtypes import DType
    from narwhals.typing import DTypeBackend, IntoDType, TimeUnit, _1DArray

    ExprT = TypeVar("ExprT", bound=PandasLikeExpr)
    UnitCurrent: TypeAlias = TimeUnit
    UnitTarget: TypeAlias = TimeUnit
    BinOpBroadcast: TypeAlias = Callable[[Any, int], Any]
    IntoRhs: TypeAlias = int


PANDAS_LIKE_IMPLEMENTATION = {
    Implementation.PANDAS,
    Implementation.CUDF,
    Implementation.MODIN,
}
PD_DATETIME_RGX = r"""^
    datetime64\[
        (?P<time_unit>s|ms|us|ns)                 # Match time unit: s, ms, us, or ns
        (?:,                                      # Begin non-capturing group for optional timezone
            \s*                                   # Optional whitespace after comma
            (?P<time_zone>                        # Start named group for timezone
                [a-zA-Z\/]+                       # Match timezone name, e.g., UTC, America/New_York
                (?:[+-]\d{2}:\d{2})?              # Optional offset in format +HH:MM or -HH:MM
                |                                 # OR
                pytz\.FixedOffset\(\d+\)          # Match pytz.FixedOffset with integer offset in parentheses
            )                                     # End time_zone group
        )?                                        # End optional timezone group
    \]                                            # Closing bracket for datetime64
$"""
PATTERN_PD_DATETIME = re.compile(PD_DATETIME_RGX, re.VERBOSE)
PD_DURATION_RGX = r"""^
    timedelta64\[
        (?P<time_unit>s|ms|us|ns)                 # Match time unit: s, ms, us, or ns
    \]                                            # Closing bracket for timedelta64
$"""
PATTERN_PD_DURATION = re.compile(PD_DURATION_RGX, re.VERBOSE)

NativeIntervalUnit: TypeAlias = Literal[
    "year",
    "quarter",
    "month",
    "week",
    "day",
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "nanosecond",
]
ALIAS_DICT = {"d": "D", "m": "min"}
UNITS_DICT: Mapping[IntervalUnit, NativeIntervalUnit] = {
    "y": "year",
    "q": "quarter",
    "mo": "month",
    "d": "day",
    "h": "hour",
    "m": "minute",
    "s": "second",
    "ms": "millisecond",
    "us": "microsecond",
    "ns": "nanosecond",
}

PANDAS_VERSION = Implementation.PANDAS._backend_version()
"""Static backend version for `pandas`.

Always available if we reached here, due to a module-level import.
"""

NUMPY_VERSION = parse_version(np)
"""Static version for `numpy`.

Always available if we reached here, as imported in both _pandas_like/dataframe.py and
_pandas_like/series.py.
"""


def is_pandas_or_modin(implementation: Implementation) -> bool:
    return implementation in {Implementation.PANDAS, Implementation.MODIN}


def align_and_extract_native(
    lhs: PandasLikeSeries, rhs: PandasLikeSeries | object
) -> tuple[pd.Series[Any], pd.Series[Any] | object]:
    """Validate RHS of binary operation.

    If the comparison isn't supported, return `NotImplemented` so that the
    "right-hand-side" operation (e.g. `__radd__`) can be tried.
    """
    from narwhals._pandas_like.series import PandasLikeSeries

    lhs_index = lhs.native.index

    if lhs._broadcast and isinstance(rhs, PandasLikeSeries) and not rhs._broadcast:
        return lhs.native.iloc[0], rhs.native

    if isinstance(rhs, PandasLikeSeries):
        if rhs._broadcast:
            return (lhs.native, rhs.native.iloc[0])
        if rhs.native.index is not lhs_index:
            return (
                lhs.native,
                set_index(rhs.native, lhs_index, implementation=rhs._implementation),
            )
        return (lhs.native, rhs.native)

    if isinstance(rhs, list):
        msg = "Expected Series or scalar, got list."
        raise TypeError(msg)

    # `rhs` must be scalar, so just leave it as-is
    return lhs.native, rhs


def set_index(
    obj: NativeNDFrameT, index: Any, *, implementation: Implementation
) -> NativeNDFrameT:
    """Wrapper around pandas' set_axis to set object index.

    We can set `copy` / `inplace` based on implementation/version.
    """
    if isinstance(index, implementation.to_native_namespace().Index) and (
        expected_len := len(index)
    ) != (actual_len := len(obj)):
        msg = f"Expected object of length {expected_len}, got length: {actual_len}"
        raise ShapeError(msg)
    if implementation is Implementation.CUDF:
        obj = obj.copy(deep=False)
        obj.index = index
        return obj
    if implementation is Implementation.PANDAS and (
        (1, 5) <= implementation._backend_version() < (3,)
    ):  # pragma: no cover
        return obj.set_axis(index, axis=0, copy=False)
    return obj.set_axis(index, axis=0)  # pragma: no cover


def rename(
    obj: NativeNDFrameT, *args: Any, implementation: Implementation, **kwargs: Any
) -> NativeNDFrameT:
    """Wrapper around pandas' rename so that we can set `copy` based on implementation/version."""
    if implementation is Implementation.PANDAS and (
        implementation._backend_version() >= (3,)
    ):  # pragma: no cover
        result = obj.rename(*args, **kwargs, inplace=False)
    else:  # pragma: no cover
        result = obj.rename(*args, **kwargs, copy=False, inplace=False)
    return cast("NativeNDFrameT", result)  # type: ignore[redundant-cast]


@functools.lru_cache(maxsize=16)
def is_dtype_non_pyarrow_string(native_dtype: Any) -> bool:
    """*There is no problem which can't be solved by adding an extra string type* pandas."""
    # TODO @dangotbanned: Investigate how we could handle `cudf` without `str(native_dtype)`
    # https://github.com/rapidsai/cudf/blob/a32b8cf62c9b086b645b0825b78b99f065b1887f/python/cudf/cudf/utils/dtypes.py#L646-L670
    return isinstance(native_dtype, pd.StringDtype) or str(native_dtype) in {
        "string",
        "string[python]",
        "string[pyarrow_numpy]",
        "<StringDtype(na_value=nan)>",  # why? why? why?
        "str",
    }


@functools.lru_cache(maxsize=16)
def non_object_native_to_narwhals_dtype(native_dtype: Any, version: Version) -> DType:  # noqa: C901, PLR0912
    dtype = str(native_dtype)

    dtypes = version.dtypes
    if dtype in {"int64", "Int64"}:
        return dtypes.Int64()
    if dtype in {"int32", "Int32"}:
        return dtypes.Int32()
    if dtype in {"int16", "Int16"}:
        return dtypes.Int16()
    if dtype in {"int8", "Int8"}:
        return dtypes.Int8()
    if dtype in {"uint64", "UInt64"}:
        return dtypes.UInt64()
    if dtype in {"uint32", "UInt32"}:
        return dtypes.UInt32()
    if dtype in {"uint16", "UInt16"}:
        return dtypes.UInt16()
    if dtype in {"uint8", "UInt8"}:
        return dtypes.UInt8()
    if dtype in {"float64", "Float64"}:
        return dtypes.Float64()
    if dtype in {"float32", "Float32"}:
        return dtypes.Float32()
    if is_dtype_non_pyarrow_string(native_dtype):
        return dtypes.String()
    if dtype in {"bool", "boolean"}:
        return dtypes.Boolean()
    if match_ := PATTERN_PD_DATETIME.match(dtype):
        dt_time_unit: TimeUnit = match_.group("time_unit")  # type: ignore[assignment]
        dt_time_zone: str | None = match_.group("time_zone")
        return dtypes.Datetime(dt_time_unit, dt_time_zone)
    if match_ := PATTERN_PD_DURATION.match(dtype):
        du_time_unit: TimeUnit = match_.group("time_unit")  # type: ignore[assignment]
        return dtypes.Duration(du_time_unit)
    return dtypes.Unknown()  # pragma: no cover


def object_native_to_narwhals_dtype(
    series: PandasLikeSeries | None, version: Version, implementation: Implementation
) -> DType:
    dtypes = version.dtypes
    if implementation is Implementation.CUDF:
        # Per conversations with their maintainers, they don't support arbitrary
        # objects, so we can just return String.
        return dtypes.String()

    infer = pd.api.types.infer_dtype
    # Arbitrary limit of 100 elements to use to sniff dtype.
    inferred_dtype = "empty" if series is None else infer(series.head(100), skipna=True)
    if inferred_dtype == "string":
        return dtypes.String()
    if inferred_dtype == "empty" and version is not Version.V1:
        # Default to String for empty Series.
        return dtypes.String()
    if inferred_dtype == "empty":
        # But preserve returning Object in V1.
        return dtypes.Object()
    return dtypes.Object()


def native_categorical_to_narwhals_dtype(
    native_dtype: pd.CategoricalDtype, version: Version, implementation: Implementation
) -> DType:
    dtypes = version.dtypes
    if version is Version.V1:
        return dtypes.Categorical()
    if native_dtype.ordered:
        into_iter = (
            _cudf_categorical_to_list(native_dtype)
            if implementation is Implementation.CUDF
            else native_dtype.categories.to_list
        )
        return dtypes.Enum(_DeferredIterable(into_iter))
    return dtypes.Categorical()


def _cudf_categorical_to_list(
    native_dtype: Any,
) -> Callable[[], list[Any]]:  # pragma: no cover
    # NOTE: https://docs.rapids.ai/api/cudf/stable/user_guide/api_docs/api/cudf.core.dtypes.categoricaldtype/#cudf.core.dtypes.CategoricalDtype
    # https://github.com/rapidsai/cudf/issues/18536
    # https://github.com/rapidsai/cudf/issues/14027
    def fn() -> list[Any]:
        return native_dtype.categories.to_arrow().to_pylist()

    return fn


CUDF_BASE_DTYPE_PREFIX = ("list", "struct", "decimal")


def native_to_narwhals_dtype(
    native_dtype: Any,
    version: Version,
    implementation: Implementation,
    *,
    allow_object: bool = False,
) -> DType:
    str_dtype = str(native_dtype)

    if is_dtype_pyarrow(native_dtype) or str_dtype.startswith(CUDF_BASE_DTYPE_PREFIX):
        from narwhals._arrow.utils import (
            native_to_narwhals_dtype as arrow_native_to_narwhals_dtype,
        )

        if hasattr(native_dtype, "to_arrow"):  # pragma: no cover
            pa_dtype: pa.DataType = native_dtype.to_arrow()  # pyright: ignore[reportAttributeAccessIssue]
        else:
            pa_dtype = native_dtype.pyarrow_dtype
        return arrow_native_to_narwhals_dtype(pa_dtype, version)
    if str_dtype == "category":
        return native_categorical_to_narwhals_dtype(native_dtype, version, implementation)
    if str_dtype != "object":
        return non_object_native_to_narwhals_dtype(native_dtype, version)
    if implementation is Implementation.DASK:
        # Per conversations with their maintainers, they don't support arbitrary
        # objects, so we can just return String.
        return version.dtypes.String()
    if allow_object:  # pragma: no cover
        return object_native_to_narwhals_dtype(None, version, implementation)
    msg = (
        "Unreachable code, object dtype should be handled separately"  # pragma: no cover
    )
    raise AssertionError(msg)


if Implementation.PANDAS._backend_version() >= (1, 2):

    def is_dtype_numpy_nullable(dtype: Any) -> TypeIs[BaseMaskedDtype]:
        """Return `True` if `dtype` is `"numpy_nullable"`."""
        # NOTE: We need a sentinel as the positive case is `BaseMaskedDtype.base = None`
        # See https://github.com/narwhals-dev/narwhals/pull/2740#discussion_r2171667055
        sentinel = object()
        return (
            isinstance(dtype, pd.api.extensions.ExtensionDtype)
            and getattr(dtype, "base", sentinel) is None
        )
else:  # pragma: no cover

    def is_dtype_numpy_nullable(dtype: Any) -> TypeIs[BaseMaskedDtype]:
        # NOTE: `base` attribute was added between 1.1-1.2
        # Checking by isinstance requires using an import path that is no longer valid
        # `1.1`: https://github.com/pandas-dev/pandas/blob/b5958ee1999e9aead1938c0bba2b674378807b3d/pandas/core/arrays/masked.py#L37
        # `1.2`: https://github.com/pandas-dev/pandas/blob/7c48ff4409c622c582c56a5702373f726de08e96/pandas/core/arrays/masked.py#L41
        # `1.5`: https://github.com/pandas-dev/pandas/blob/35b0d1dcadf9d60722c055ee37442dc76a29e64c/pandas/core/dtypes/dtypes.py#L1609
        if isinstance(dtype, pd.api.extensions.ExtensionDtype):
            from pandas.core.arrays.masked import (  # type: ignore[attr-defined]
                BaseMaskedDtype as OldBaseMaskedDtype,  # pyright: ignore[reportAttributeAccessIssue]
            )

            return isinstance(dtype, OldBaseMaskedDtype)
        return False


def get_dtype_backend(dtype: Any, implementation: Implementation) -> DTypeBackend:
    """Get dtype backend for pandas type.

    Matches pandas' `dtype_backend` argument in `convert_dtypes`.
    """
    if implementation is Implementation.CUDF:
        return None
    if is_dtype_pyarrow(dtype):
        return "pyarrow"
    return "numpy_nullable" if is_dtype_numpy_nullable(dtype) else None


# NOTE: Use this to avoid annotating inline
def iter_dtype_backends(
    dtypes: Iterable[Any], implementation: Implementation
) -> Iterator[DTypeBackend]:
    """Yield a `DTypeBackend` per-dtype.

    Matches pandas' `dtype_backend` argument in `convert_dtypes`.
    """
    return (get_dtype_backend(dtype, implementation) for dtype in dtypes)


@functools.lru_cache(maxsize=16)
def is_dtype_pyarrow(dtype: Any) -> TypeIs[pd.ArrowDtype]:
    return hasattr(pd, "ArrowDtype") and isinstance(dtype, pd.ArrowDtype)


dtypes = Version.MAIN.dtypes
NW_TO_PD_DTYPES_INVARIANT: Mapping[type[DType], str] = {
    # TODO(Unassigned): is there no pyarrow-backed categorical?
    # or at least, convert_dtypes(dtype_backend='pyarrow') doesn't
    # convert to it?
    dtypes.Categorical: "category",
    dtypes.Object: "object",
}
NW_TO_PD_DTYPES_BACKEND: Mapping[type[DType], Mapping[DTypeBackend, str | type[Any]]] = {
    dtypes.Float64: {
        "pyarrow": "Float64[pyarrow]",
        "numpy_nullable": "Float64",
        None: "float64",
    },
    dtypes.Float32: {
        "pyarrow": "Float32[pyarrow]",
        "numpy_nullable": "Float32",
        None: "float32",
    },
    dtypes.Int64: {"pyarrow": "Int64[pyarrow]", "numpy_nullable": "Int64", None: "int64"},
    dtypes.Int32: {"pyarrow": "Int32[pyarrow]", "numpy_nullable": "Int32", None: "int32"},
    dtypes.Int16: {"pyarrow": "Int16[pyarrow]", "numpy_nullable": "Int16", None: "int16"},
    dtypes.Int8: {"pyarrow": "Int8[pyarrow]", "numpy_nullable": "Int8", None: "int8"},
    dtypes.UInt64: {
        "pyarrow": "UInt64[pyarrow]",
        "numpy_nullable": "UInt64",
        None: "uint64",
    },
    dtypes.UInt32: {
        "pyarrow": "UInt32[pyarrow]",
        "numpy_nullable": "UInt32",
        None: "uint32",
    },
    dtypes.UInt16: {
        "pyarrow": "UInt16[pyarrow]",
        "numpy_nullable": "UInt16",
        None: "uint16",
    },
    dtypes.UInt8: {"pyarrow": "UInt8[pyarrow]", "numpy_nullable": "UInt8", None: "uint8"},
    dtypes.Boolean: {
        "pyarrow": "boolean[pyarrow]",
        "numpy_nullable": "boolean",
        None: "bool",
    },
}


def narwhals_to_native_dtype(  # noqa: C901, PLR0912
    dtype: IntoDType,
    dtype_backend: DTypeBackend,
    implementation: Implementation,
    version: Version,
) -> str | PandasDtype:
    if dtype_backend not in {None, "pyarrow", "numpy_nullable"}:
        msg = f"Expected one of {{None, 'pyarrow', 'numpy_nullable'}}, got: '{dtype_backend}'"
        raise ValueError(msg)
    dtypes = version.dtypes
    base_type = dtype.base_type()
    if pd_type := NW_TO_PD_DTYPES_INVARIANT.get(base_type):
        return pd_type
    if into_pd_type := NW_TO_PD_DTYPES_BACKEND.get(base_type):
        return into_pd_type[dtype_backend]
    if issubclass(base_type, dtypes.String):
        if dtype_backend == "pyarrow":
            import pyarrow as pa  # ignore-banned-import

            # Note: this is different from `string[pyarrow]`, even though the repr
            # looks the same.
            # >>> pd.DataFrame({'a':['foo']}, dtype='string[pyarrow]')['a'].str.len()
            # 0    3
            # Name: a, dtype: Int64
            # >>> pd.DataFrame({'a':['foo']}, dtype=pd.ArrowDtype(pa.string()))['a'].str.len()
            # 0    3
            # Name: a, dtype: int32[pyarrow]
            #
            # `ArrowDType(pa.string())` is what `.convert_dtypes(dtype_backend='pyarrow')` converts to,
            # so we use that here.
            return pd.ArrowDtype(pa.string())
        if dtype_backend == "numpy_nullable":
            return "string"
        return str
    if isinstance_or_issubclass(dtype, dtypes.Datetime):
        if is_pandas_or_modin(implementation) and PANDAS_VERSION < (
            2,
        ):  # pragma: no cover
            if isinstance(dtype, dtypes.Datetime) and dtype.time_unit != "ns":
                found = requires._unparse_version(PANDAS_VERSION)
                available = f"available in 'pandas>=2.0', found version {found!r}."
                changelog_url = "https://pandas.pydata.org/docs/dev/whatsnew/v2.0.0.html#construction-with-datetime64-or-timedelta64-dtype-with-unsupported-resolution"
                msg = (
                    f"`nw.Datetime(time_unit={dtype.time_unit!r})` is only {available}\n"
                    "Narwhals has fallen back to using `time_unit='ns'` to avoid an error.\n\n"
                    "Hint: to avoid this warning, consider either:\n"
                    f"- Upgrading pandas: {changelog_url}\n"
                    f"- Using a bare `nw.Datetime`, if this precision is not important"
                )
                issue_warning(msg, UserWarning)
            dt_time_unit = "ns"
        else:
            dt_time_unit = dtype.time_unit

        if dtype_backend == "pyarrow":
            tz_part = f", tz={tz}" if (tz := dtype.time_zone) else ""
            return f"timestamp[{dt_time_unit}{tz_part}][pyarrow]"
        tz_part = f", {tz}" if (tz := dtype.time_zone) else ""
        return f"datetime64[{dt_time_unit}{tz_part}]"
    if isinstance_or_issubclass(dtype, dtypes.Duration):
        if is_pandas_or_modin(implementation) and PANDAS_VERSION < (
            2,
        ):  # pragma: no cover
            du_time_unit = "ns"
        else:
            du_time_unit = dtype.time_unit
        return (
            f"duration[{du_time_unit}][pyarrow]"
            if dtype_backend == "pyarrow"
            else f"timedelta64[{du_time_unit}]"
        )
    if isinstance_or_issubclass(dtype, dtypes.Date):
        try:
            import pyarrow as pa  # ignore-banned-import
        except ModuleNotFoundError as exc:
            # BUG: Never re-raised?
            msg = "'pyarrow>=13.0.0' is required for `Date` dtype."
            raise ModuleNotFoundError(msg) from exc
        return "date32[pyarrow]"
    if isinstance_or_issubclass(dtype, dtypes.Enum):
        if version is Version.V1:
            msg = "Converting to Enum is not supported in narwhals.stable.v1"
            raise NotImplementedError(msg)
        if isinstance(dtype, dtypes.Enum):
            ns = implementation.to_native_namespace()
            return ns.CategoricalDtype(dtype.categories, ordered=True)
        msg = "Can not cast / initialize Enum without categories present"
        raise ValueError(msg)
    if issubclass(
        base_type,
        (
            dtypes.Struct,
            dtypes.Array,
            dtypes.List,
            dtypes.Time,
            dtypes.Binary,
            dtypes.Decimal,
        ),
    ):
        return narwhals_to_native_arrow_dtype(dtype, implementation, version)
    msg = f"Unknown dtype: {dtype}"  # pragma: no cover
    raise AssertionError(msg)


def narwhals_to_native_arrow_dtype(
    dtype: IntoDType, implementation: Implementation, version: Version
) -> pd.ArrowDtype:
    if is_pandas_or_modin(implementation) and PANDAS_VERSION >= (2, 2):
        try:
            import pyarrow as pa  # ignore-banned-import  # noqa: F401
        except ImportError as exc:  # pragma: no cover
            msg = (
                f"Unable to convert to {dtype} due to the following exception: {exc.msg}"
            )
            raise ImportError(msg) from exc
        from narwhals._arrow.utils import narwhals_to_native_dtype as _to_arrow_dtype

        return pd.ArrowDtype(_to_arrow_dtype(dtype, version))
    msg = (  # pragma: no cover
        f"Converting to {dtype} dtype is not supported for implementation "
        f"{implementation} and version {version}."
    )
    raise NotImplementedError(msg)


def int_dtype_mapper(dtype: Any) -> str:
    if "pyarrow" in str(dtype):
        return "Int64[pyarrow]"
    if str(dtype).lower() != str(dtype):  # pragma: no cover
        return "Int64"
    return "int64"


_TIMESTAMP_DATETIME_OP_FACTOR: Mapping[
    tuple[UnitCurrent, UnitTarget], tuple[BinOpBroadcast, IntoRhs]
] = {
    ("ns", "us"): (operator.floordiv, 1_000),
    ("ns", "ms"): (operator.floordiv, 1_000_000),
    ("us", "ns"): (operator.mul, NS_PER_MICROSECOND),
    ("us", "ms"): (operator.floordiv, 1_000),
    ("ms", "ns"): (operator.mul, NS_PER_MILLISECOND),
    ("ms", "us"): (operator.mul, 1_000),
    ("s", "ns"): (operator.mul, NS_PER_SECOND),
    ("s", "us"): (operator.mul, US_PER_SECOND),
    ("s", "ms"): (operator.mul, MS_PER_SECOND),
}


def calculate_timestamp_datetime(
    s: NativeSeriesT, current: TimeUnit, time_unit: TimeUnit
) -> NativeSeriesT:
    if current == time_unit:
        return s
    if item := _TIMESTAMP_DATETIME_OP_FACTOR.get((current, time_unit)):
        fn, factor = item
        return fn(s, factor)
    msg = (  # pragma: no cover
        f"unexpected time unit {current}, please report an issue at "
        "https://github.com/narwhals-dev/narwhals"
    )
    raise AssertionError(msg)


_TIMESTAMP_DATE_FACTOR: Mapping[TimeUnit, int] = {
    "ns": NS_PER_SECOND,
    "us": US_PER_SECOND,
    "ms": MS_PER_SECOND,
    "s": 1,
}


def calculate_timestamp_date(s: NativeSeriesT, time_unit: TimeUnit) -> NativeSeriesT:
    return s * SECONDS_PER_DAY * _TIMESTAMP_DATE_FACTOR[time_unit]


def select_columns_by_name(
    df: NativeDataFrameT,
    column_names: list[str] | _1DArray,  # NOTE: Cannot be a tuple!
    implementation: Implementation,
) -> NativeDataFrameT | Any:
    """Select columns by name.

    Prefer this over `df.loc[:, column_names]` as it's
    generally more performant.
    """
    if len(column_names) == df.shape[1] and (df.columns == column_names).all():
        return df
    if (df.columns.dtype.kind == "b") or (
        implementation is Implementation.PANDAS
        and implementation._backend_version() < (1, 5)
    ):
        # See https://github.com/narwhals-dev/narwhals/issues/1349#issuecomment-2470118122
        # for why we need this
        if error := check_columns_exist(column_names, available=df.columns.tolist()):
            raise error
        return df.loc[:, column_names]
    try:
        return df[column_names]
    except KeyError as e:
        if error := check_columns_exist(column_names, available=df.columns.tolist()):
            raise error from e
        raise


def is_non_nullable_boolean(s: PandasLikeSeries) -> bool:
    # cuDF booleans are nullable but the native dtype is still 'bool'.
    return (
        s._implementation
        in {Implementation.PANDAS, Implementation.MODIN, Implementation.DASK}
        and s.native.dtype == "bool"
    )


def import_array_module(implementation: Implementation, /) -> ModuleType:
    """Returns numpy or cupy module depending on the given implementation."""
    if implementation in {Implementation.PANDAS, Implementation.MODIN}:
        import numpy as np

        return np
    if implementation is Implementation.CUDF:
        import cupy as cp  # ignore-banned-import  # cuDF dependency.

        return cp
    msg = f"Expected pandas/modin/cudf, got: {implementation}"  # pragma: no cover
    raise AssertionError(msg)


class PandasLikeSeriesNamespace(EagerSeriesNamespace["PandasLikeSeries", Any]): ...


def make_group_by_kwargs(*, drop_null_keys: bool) -> dict[str, bool]:
    return {"sort": False, "as_index": True, "dropna": drop_null_keys, "observed": True}


def broadcast_series_to_index(
    native: pd.Series[Any],
    index: Any,
    *,
    is_nested: bool,
    series_class: type[pd.Series[Any]],
) -> pd.Series[Any]:
    """Broadcast a scalar value from a (one element) Series to match a target index.

    For nested (arrow-backed) types, we rely on
    [`pandas.array`](https://pandas.pydata.org/docs/reference/api/pandas.array.html).

    Arguments:
        native: The native pandas-like Series containing the scalar value to broadcast.
        index: The target index to broadcast to.
        is_nested: Whether the Series has a nested (arrow-backed) dtype.
        series_class: Series class to use for constructing the result.

    Returns:
        A new Series with the scalar value broadcast to match the target index.
    """
    value = native.iloc[0]
    if is_nested:
        from narwhals._arrow.utils import repeat

        # NOTE: Ignore typing because `pandas-stubs` are wrong
        # TODO(FBruzzesi): Should we pass the `copy=False` flag?
        pa_array = pd.array(repeat(value, len(index)), dtype=native.dtype)  # type: ignore[arg-type]

        return series_class(pa_array, index=index, name=native.name)

    return series_class(value, index=index, dtype=native.dtype, name=native.name)



from __future__ import annotations

import operator
from datetime import date, datetime
from functools import reduce
from itertools import chain
from typing import TYPE_CHECKING, cast

import dask.dataframe as dd
import pandas as pd

from narwhals._compliant import DepthTrackingNamespace, LazyNamespace
from narwhals._dask.dataframe import DaskLazyFrame
from narwhals._dask.expr import DaskExpr
from narwhals._dask.selectors import DaskSelectorNamespace
from narwhals._dask.utils import (
    align_series_full_broadcast,
    narwhals_to_native_dtype,
    validate_comparand,
)
from narwhals._expression_parsing import (
    combine_alias_output_names,
    combine_evaluate_output_names,
)
from narwhals._utils import Implementation, is_nested_literal, not_implemented, zip_strict

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    import dask.dataframe.dask_expr as dx

    from narwhals._utils import Version
    from narwhals.typing import (
        ConcatMethod,
        CorrelationMethod,
        IntoDType,
        NonNestedLiteral,
    )


class DaskNamespace(
    LazyNamespace[DaskLazyFrame, DaskExpr, dd.DataFrame],
    DepthTrackingNamespace[DaskLazyFrame, DaskExpr],
):
    _implementation: Implementation = Implementation.DASK

    @property
    def selectors(self) -> DaskSelectorNamespace:
        return DaskSelectorNamespace.from_namespace(self)

    @property
    def _expr(self) -> type[DaskExpr]:
        return DaskExpr

    @property
    def _lazyframe(self) -> type[DaskLazyFrame]:
        return DaskLazyFrame

    def __init__(self, *, version: Version) -> None:
        self._version = version

    def lit(self, value: NonNestedLiteral, dtype: IntoDType | None) -> DaskExpr:
        if is_nested_literal(value):
            msg = f"Nested structures are not supported for Dask backend, found {type(value).__name__}"
            raise NotImplementedError(msg)

        def func(df: DaskLazyFrame) -> list[dx.Series]:
            if dtype is not None:
                native_dtype = narwhals_to_native_dtype(dtype, self._version)
                native_pd_series = pd.Series([value], dtype=native_dtype, name="literal")
            elif isinstance(value, date) and not isinstance(
                value, datetime
            ):  # pragma: no cover
                # Dask auto-infers this as object type, which causes issues down the line.
                # This shows up in TPC-H q8.
                native_dtype = "date32[pyarrow]"
                native_pd_series = pd.Series([value], dtype=native_dtype, name="literal")
            else:
                native_pd_series = pd.Series([value], name="literal")
            npartitions = df._native_frame.npartitions
            dask_series = dd.from_pandas(native_pd_series, npartitions=npartitions)
            return [dask_series[0].to_series()]

        return self._expr(
            func,
            evaluate_output_names=lambda _df: ["literal"],
            alias_output_names=None,
            version=self._version,
        )

    def len(self) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            # We don't allow dataframes with 0 columns, so `[0]` is safe.
            return [df._native_frame[df.columns[0]].size.to_series()]

        return self._expr(
            func,
            evaluate_output_names=lambda _df: ["len"],
            alias_output_names=None,
            version=self._version,
        )

    def all_horizontal(self, *exprs: DaskExpr, ignore_nulls: bool) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series: Iterator[dx.Series] = chain.from_iterable(e(df) for e in exprs)
            # Note on `ignore_nulls`: Dask doesn't support storing arbitrary Python
            # objects in `object` dtype, so we don't need the same check we have for pandas-like.
            if ignore_nulls:
                # NumPy-backed 'bool' dtype can't contain nulls so doesn't need filling.
                series = (s if s.dtype == "bool" else s.fillna(True) for s in series)
            return [reduce(operator.and_, align_series_full_broadcast(df, *series))]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def any_horizontal(self, *exprs: DaskExpr, ignore_nulls: bool) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series: Iterator[dx.Series] = chain.from_iterable(e(df) for e in exprs)
            if ignore_nulls:
                series = (s if s.dtype == "bool" else s.fillna(False) for s in series)
            return [reduce(operator.or_, align_series_full_broadcast(df, *series))]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def sum_horizontal(self, *exprs: DaskExpr) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series = align_series_full_broadcast(
                df, *(s for _expr in exprs for s in _expr(df))
            )
            return [dd.concat(series, axis=1).sum(axis=1)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def concat(
        self, items: Iterable[DaskLazyFrame], *, how: ConcatMethod
    ) -> DaskLazyFrame:
        if not items:
            msg = "No items to concatenate"  # pragma: no cover
            raise AssertionError(msg)
        dfs = [i._native_frame for i in items]
        cols_0 = dfs[0].columns
        if how == "vertical":
            for i, df in enumerate(dfs[1:], start=1):
                cols_current = df.columns
                if not (
                    (len(cols_current) == len(cols_0)) and (cols_current == cols_0).all()
                ):
                    msg = (
                        "unable to vstack, column names don't match:\n"
                        f"   - dataframe 0: {cols_0.to_list()}\n"
                        f"   - dataframe {i}: {cols_current.to_list()}\n"
                    )
                    raise TypeError(msg)
            return DaskLazyFrame(
                dd.concat(dfs, axis=0, join="inner"), version=self._version
            )
        if how == "diagonal":
            return DaskLazyFrame(
                dd.concat(dfs, axis=0, join="outer"), version=self._version
            )

        raise NotImplementedError

    def mean_horizontal(self, *exprs: DaskExpr) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            expr_results = [s for _expr in exprs for s in _expr(df)]
            series = align_series_full_broadcast(df, *(s.fillna(0) for s in expr_results))
            non_na = align_series_full_broadcast(
                df, *(1 - s.isna() for s in expr_results)
            )
            num = reduce(lambda x, y: x + y, series)  # pyright: ignore[reportOperatorIssue]
            den = reduce(lambda x, y: x + y, non_na)  # pyright: ignore[reportOperatorIssue]
            return [cast("dx.Series", num / den)]  # pyright: ignore[reportOperatorIssue]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def min_horizontal(self, *exprs: DaskExpr) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series = align_series_full_broadcast(
                df, *(s for _expr in exprs for s in _expr(df))
            )

            return [dd.concat(series, axis=1).min(axis=1)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def max_horizontal(self, *exprs: DaskExpr) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series = align_series_full_broadcast(
                df, *(s for _expr in exprs for s in _expr(df))
            )

            return [dd.concat(series, axis=1).max(axis=1)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def concat_str(
        self, *exprs: DaskExpr, separator: str, ignore_nulls: bool
    ) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            expr_results = [s for _expr in exprs for s in _expr(df)]
            series = (
                s.astype(str) for s in align_series_full_broadcast(df, *expr_results)
            )
            null_mask = [s.isna() for s in align_series_full_broadcast(df, *expr_results)]

            if not ignore_nulls:
                null_mask_result = reduce(operator.or_, null_mask)
                result = reduce(lambda x, y: x + separator + y, series).where(
                    ~null_mask_result, None
                )
            else:
                init_value, *values = [
                    s.where(~nm, "") for s, nm in zip_strict(series, null_mask)
                ]

                separators = (
                    nm.map({True: "", False: separator}, meta=str)
                    for nm in null_mask[:-1]
                )
                result = reduce(
                    operator.add,
                    (s + v for s, v in zip_strict(separators, values)),
                    init_value,
                )

            return [result]

        return self._expr(
            call=func,
            evaluate_output_names=getattr(
                exprs[0], "_evaluate_output_names", lambda _df: ["literal"]
            ),
            alias_output_names=getattr(exprs[0], "_alias_output_names", None),
            version=self._version,
        )

    def coalesce(self, *exprs: DaskExpr) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            series = align_series_full_broadcast(
                df, *(s for _expr in exprs for s in _expr(df))
            )
            return [reduce(lambda x, y: x.fillna(y), series)]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(*exprs),
            alias_output_names=combine_alias_output_names(*exprs),
            version=self._version,
        )

    def when_then(
        self, predicate: DaskExpr, then: DaskExpr, otherwise: DaskExpr | None = None
    ) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            then_value = df._evaluate_single_output_expr(then)
            otherwise_value = (
                df._evaluate_single_output_expr(otherwise)
                if otherwise is not None
                else otherwise
            )

            condition = df._evaluate_single_output_expr(predicate)
            # re-evaluate DataFrame if the condition aggregates to force
            # then/otherwise to be evaluated against the aggregated frame
            if all(
                x._metadata.is_scalar_like
                for x in (
                    (predicate, then)
                    if otherwise is None
                    else (predicate, then, otherwise)
                )
            ):
                new_df = df._with_native(condition.to_frame())
                condition = df._evaluate_single_output_expr(predicate.broadcast())
                df = new_df

            if otherwise is None:
                (condition, then_series) = align_series_full_broadcast(
                    df, condition, then_value
                )
                validate_comparand(condition, then_series)
                return [then_series.where(condition)]  # pyright: ignore[reportArgumentType]
            (condition, then_series, otherwise_series) = align_series_full_broadcast(
                df, condition, then_value, otherwise_value
            )
            validate_comparand(condition, then_series)
            validate_comparand(condition, otherwise_series)
            return [then_series.where(condition, otherwise_series)]  # pyright: ignore[reportArgumentType]

        return self._expr(
            call=func,
            evaluate_output_names=getattr(
                then, "_evaluate_output_names", lambda _df: ["literal"]
            ),
            alias_output_names=getattr(then, "_alias_output_names", None),
            version=self._version,
        )

    def corr(self, a: DaskExpr, b: DaskExpr, *, method: CorrelationMethod) -> DaskExpr:
        def func(df: DaskLazyFrame) -> list[dx.Series]:
            a_ = df._evaluate_single_output_expr(a)
            b_ = df._evaluate_single_output_expr(b)
            return [a_.corr(b_, method=method).to_series()]

        return self._expr(
            call=func,
            evaluate_output_names=combine_evaluate_output_names(a, b),
            alias_output_names=combine_alias_output_names(a, b),
            version=self._version,
        )

    struct = not_implemented()



from __future__ import annotations

import datetime as dt
import os
import re
from decimal import Decimal
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals.exceptions import DuplicateError, InvalidOperationError
from tests.utils import (
    DUCKDB_VERSION,
    PANDAS_VERSION,
    POLARS_VERSION,
    PYARROW_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from narwhals.typing import NonNestedLiteral


data: Mapping[str, Any] = {"a": [1, 1, 3], "b": [4, 4, 6], "c": [7.0, 8.0, 9.0]}

POLARS_COLLECT_STREAMING_ENGINE = os.environ.get("NARWHALS_POLARS_NEW_STREAMING", None)


def test_group_by_complex() -> None:
    pytest.importorskip("pandas")
    pytest.importorskip("pyarrow")
    import pandas as pd
    import pyarrow as pa

    expected = {"a": [1, 3], "b": [-3.5, -3.0]}

    df = nw.from_native(pd.DataFrame(data))
    with pytest.warns(UserWarning, match="complex group-by"):
        result_pd = nw.to_native(
            df.group_by("a").agg((nw.col("b") - nw.col("c").mean()).mean()).sort("a")
        )
    assert_equal_data(result_pd, expected)
    with pytest.raises(ValueError, match="complex aggregation"):
        nw.from_native(pa.table({"a": [1, 1, 2], "b": [4, 5, 6]})).group_by("a").agg(
            (nw.col("b") - nw.col("c").mean()).mean()
        )


def test_group_by_complex_polars() -> None:
    pytest.importorskip("polars")
    import polars as pl

    expected = {"a": [1, 3], "b": [-3.5, -3.0]}

    df_lazy = pl.LazyFrame(data)
    lf = nw.from_native(df_lazy).lazy()
    result_pl = lf.group_by("a").agg((nw.col("b") - nw.col("c").mean()).mean()).sort("a")
    assert_equal_data(result_pl, expected)


def test_invalid_group_by_dask() -> None:
    pytest.importorskip("dask")
    import dask.dataframe as dd

    df_dask = dd.from_dict(data, npartitions=1)

    with pytest.raises(ValueError, match=r"Non-trivial complex aggregation found"):
        nw.from_native(df_dask).group_by("a").agg(nw.col("b").abs().min())


def test_group_by_iter(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    expected_keys = [(1,), (3,)]
    keys = []
    for key, sub_df in df.group_by("a"):
        if key == (1,):
            expected = {"a": [1, 1], "b": [4, 4], "c": [7.0, 8.0]}
            assert_equal_data(sub_df, expected)
            assert isinstance(sub_df, nw.DataFrame)
        keys.append(key)
    assert sorted(keys) == sorted(expected_keys)
    expected_keys = [(1, 4), (3, 6)]  # type: ignore[list-item]
    keys = []
    for key, _ in df.group_by("a", "b"):
        keys.append(key)
    assert sorted(keys) == sorted(expected_keys)
    keys = []
    for key, _ in df.group_by(["a", "b"]):
        keys.append(key)
    assert sorted(keys) == sorted(expected_keys)


def test_group_by_iter_non_str_pandas() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    expected = {"a": {0: [1], 1: ["a"]}, "b": {0: [2], 1: ["b"]}}
    df = nw.from_native(pd.DataFrame({0: [1, 2], 1: ["a", "b"]}))
    groups: dict[Any, Any] = {keys[0]: df for keys, df in df.group_by(1)}  # type: ignore[call-overload]
    assert groups.keys() == {"a", "b"}
    groups["a"] = groups["a"].to_dict(as_series=False)
    groups["b"] = groups["b"].to_dict(as_series=False)
    assert_equal_data(groups, expected)


def test_group_by_nw_all(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 1, 2], "b": [4, 5, 6], "c": [7, 8, 9]}))
    result = df.group_by("a").agg(nw.all().sum()).sort("a")
    expected = {"a": [1, 2], "b": [9, 6], "c": [15, 9]}
    assert_equal_data(result, expected)
    result = df.group_by("a").agg(nw.all().sum().name.suffix("_sum")).sort("a")
    expected = {"a": [1, 2], "b_sum": [9, 6], "c_sum": [15, 9]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("attr", "expected"),
    [
        ("sum", {"a": [1, 2], "b": [3, 3]}),
        ("mean", {"a": [1, 2], "b": [1.5, 3]}),
        ("max", {"a": [1, 2], "b": [2, 3]}),
        ("min", {"a": [1, 2], "b": [1, 3]}),
        ("std", {"a": [1, 2], "b": [0.707107, None]}),
        ("var", {"a": [1, 2], "b": [0.5, None]}),
        ("len", {"a": [1, 2], "b": [3, 1]}),
        ("n_unique", {"a": [1, 2], "b": [3, 1]}),
        ("count", {"a": [1, 2], "b": [2, 1]}),
    ],
)
def test_group_by_depth_1_agg(
    constructor: Constructor, attr: str, expected: dict[str, list[int | float]]
) -> None:
    if "pandas_pyarrow" in str(constructor) and attr == "var" and PANDAS_VERSION < (2, 1):
        pytest.skip(
            "Known issue with variance calculation in pandas 2.0.x with pyarrow backend in groupby operations"
        )
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 1, 1, 2], "b": [1, None, 2, 3]}
    expr = getattr(nw.col("b"), attr)()
    result = nw.from_native(constructor(data)).group_by("a").agg(expr).sort("a")
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("values", "expected"),
    [
        (
            {"x": [True, True, True, False, False, False]},
            {"all": [True, False, False], "any": [True, True, False]},
        ),
        (
            {"x": [True, None, False, None, None, None]},
            {"all": [True, False, True], "any": [True, False, False]},
        ),
    ],
    ids=["not-nullable", "nullable"],
)
def test_group_by_depth_1_agg_bool_ops(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    values: dict[str, list[bool]],
    expected: dict[str, list[bool]],
) -> None:
    if ("dask-nullable" in request.node.callspec.id) or ("cudf" in str(constructor)):
        request.applymarker(pytest.mark.xfail)

    data = {"a": [1, 1, 2, 2, 3, 3], **values}
    result = (
        nw.from_native(constructor(data))
        .group_by("a")
        .agg(nw.col("x").all().alias("all"), nw.col("x").any().alias("any"))
        .sort("a")
    )
    assert_equal_data(result, {"a": [1, 2, 3], **expected})


@pytest.mark.parametrize(
    ("attr", "ddof"), [("std", 0), ("var", 0), ("std", 2), ("var", 2)]
)
def test_group_by_depth_1_std_var(constructor: Constructor, attr: str, ddof: int) -> None:
    data = {"a": [1, 1, 1, 2, 2, 2], "b": [4, 5, 6, 0, 5, 5]}
    _pow = 0.5 if attr == "std" else 1
    expected = {
        "a": [1, 2],
        "b": [
            (sum((v - 5) ** 2 for v in [4, 5, 6]) / (3 - ddof)) ** _pow,
            (sum((v - 10 / 3) ** 2 for v in [0, 5, 5]) / (3 - ddof)) ** _pow,
        ],
    }
    expr = getattr(nw.col("b"), attr)(ddof=ddof)
    result = nw.from_native(constructor(data)).group_by("a").agg(expr).sort("a")
    assert_equal_data(result, expected)


def test_group_by_median(constructor: Constructor) -> None:
    data = {"a": [1, 1, 1, 2, 2, 2], "b": [5, 4, 6, 7, 3, 2]}
    result = (
        nw.from_native(constructor(data))
        .group_by("a")
        .agg(nw.col("b").median())
        .sort("a")
    )
    expected = {"a": [1, 2], "b": [5, 3]}
    assert_equal_data(result, expected)


def test_group_by_n_unique_w_missing(constructor: Constructor) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 1, 2], "b": [4, None, 5], "c": [None, None, 7], "d": [1, 1, 3]}
    result = (
        nw.from_native(constructor(data))
        .group_by("a")
        .agg(
            nw.col("b").n_unique(),
            c_n_unique=nw.col("c").n_unique(),
            c_n_min=nw.col("b").min(),
            d_n_unique=nw.col("d").n_unique(),
        )
        .sort("a")
    )
    expected = {
        "a": [1, 2],
        "b": [2, 1],
        "c_n_unique": [1, 1],
        "c_n_min": [4, 5],
        "d_n_unique": [1, 1],
    }
    assert_equal_data(result, expected)


def test_group_by_same_name_twice() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df = pd.DataFrame({"a": [1, 1, 2], "b": [4, 5, 6]})
    pattern = re.compile(
        "expected unique.+names.+'b'.+2 times", re.IGNORECASE | re.DOTALL
    )
    with pytest.raises(DuplicateError, match=pattern):
        nw.from_native(df).group_by("a").agg(nw.col("b").sum(), nw.col("b").n_unique())


def test_group_by_empty_result_pandas() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df_any = pd.DataFrame({"a": [1, 2, 3], "b": [4, 3, 2]})
    df = nw.from_native(df_any, eager_only=True)
    with pytest.raises(ValueError, match="No results"):
        df.filter(nw.col("a") < 0).group_by("a").agg(
            nw.col("b").sum().round(2).alias("c")
        )


def test_group_by_simple_named(constructor: Constructor) -> None:
    data = {"a": [1, 1, 2], "b": [4, 5, 6], "c": [7, 2, 1]}
    df = nw.from_native(constructor(data)).lazy()
    result = (
        df.group_by("a").agg(b_min=nw.col("b").min(), b_max=nw.col("b").max()).sort("a")
    )
    expected = {"a": [1, 2], "b_min": [4, 6], "b_max": [5, 6]}
    assert_equal_data(result, expected)


def test_group_by_simple_unnamed(constructor: Constructor) -> None:
    data = {"a": [1, 1, 2], "b": [4, 5, 6], "c": [7, 2, 1]}
    df = nw.from_native(constructor(data)).lazy()
    result = df.group_by("a").agg(nw.col("b").min(), nw.col("c").max()).sort("a")
    expected = {"a": [1, 2], "b": [4, 6], "c": [7, 1]}
    assert_equal_data(result, expected)


def test_group_by_multiple_keys(constructor: Constructor) -> None:
    data = {"a": [1, 1, 2], "b": [4, 4, 6], "c": [7, 2, 1]}
    df = nw.from_native(constructor(data)).lazy()
    result = (
        df.group_by("a", "b")
        .agg(c_min=nw.col("c").min(), c_max=nw.col("c").max())
        .sort("a")
    )
    expected = {"a": [1, 2], "b": [4, 6], "c_min": [2, 1], "c_max": [7, 1]}
    assert_equal_data(result, expected)


def test_key_with_nulls(constructor: Constructor, request: pytest.FixtureRequest) -> None:
    if "modin" in str(constructor):
        request.applymarker(pytest.mark.xfail(reason="Modin flaky here", strict=False))

    data = {"b": [4, 5, None], "a": [1, 2, 3]}
    result = (
        nw.from_native(constructor(data))
        .group_by("b")
        .agg(nw.len(), nw.col("a").min())
        .sort("a")
        .with_columns(nw.col("b").cast(nw.Float64))
    )
    expected = {"b": [4.0, 5, None], "len": [1, 1, 1], "a": [1, 2, 3]}
    assert_equal_data(result, expected)


def test_key_with_nulls_ignored(constructor: Constructor) -> None:
    data = {"b": [4, 5, None], "a": [1, 2, 3]}
    result = (
        nw.from_native(constructor(data))
        .group_by("b", drop_null_keys=True)
        .agg(nw.len(), nw.col("a").min())
        .sort("a")
        .with_columns(nw.col("b").cast(nw.Float64))
    )
    expected = {"b": [4.0, 5], "len": [1, 1], "a": [1, 2]}
    assert_equal_data(result, expected)


def test_key_with_nulls_iter(constructor_eager: ConstructorEager) -> None:
    data = {
        "b": [None, "4", "5", None, "7"],
        "a": [None, 1, 2, 3, 4],
        "c": [None, "4", "3", None, None],
    }
    result = dict(
        nw.from_native(constructor_eager(data), eager_only=True)
        .group_by("b", "c", drop_null_keys=True)
        .__iter__()
    )

    assert len(result) == 2
    assert_equal_data(result[("4", "4")], {"b": ["4"], "a": [1], "c": ["4"]})
    assert_equal_data(result[("5", "3")], {"b": ["5"], "a": [2], "c": ["3"]})

    result = dict(
        nw.from_native(constructor_eager(data), eager_only=True)
        .group_by("b", "c", drop_null_keys=False)
        .__iter__()
    )
    assert_equal_data(result[("4", "4")], {"b": ["4"], "a": [1], "c": ["4"]})
    assert_equal_data(result[("5", "3")], {"b": ["5"], "a": [2], "c": ["3"]})
    assert len(result) == 4


def test_no_agg(constructor: Constructor) -> None:
    result = nw.from_native(constructor(data)).group_by(["a", "b"]).agg().sort("a", "b")

    expected = {"a": [1, 3], "b": [4, 6]}
    assert_equal_data(result, expected)


def test_group_by_categorical(constructor: Constructor) -> None:
    if any(x in str(constructor) for x in ("pyspark", "duckdb", "ibis")):
        pytest.skip(reason="no categorical support")
    if "pyarrow_table" in str(constructor) and PYARROW_VERSION < (
        15,
    ):  # pragma: no cover
        # https://github.com/narwhals-dev/narwhals/issues/1078
        pytest.skip(
            reason="The defaults for grouping by categories in pandas are different"
        )

    data = {"g1": ["a", "a", "b", "b"], "g2": ["x", "y", "x", "z"], "x": [1, 2, 3, 4]}
    df = nw.from_native(constructor(data))
    result = (
        df.with_columns(
            g1=nw.col("g1").cast(nw.Categorical()), g2=nw.col("g2").cast(nw.Categorical())
        )
        .group_by(["g1", "g2"])
        .agg(nw.col("x").sum())
        .sort("x")
    )
    assert_equal_data(result, data)


def test_group_by_shift_raises(constructor: Constructor) -> None:
    df_native = {"a": [1, 2, 3], "b": [1, 1, 2]}
    df = nw.from_native(constructor(df_native))
    with pytest.raises(InvalidOperationError, match="does not aggregate"):
        df.group_by("b").agg(nw.col("a").abs())


def test_double_same_aggregation(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("dask",)):
        # bugged in dask https://github.com/dask/dask/issues/11612
        # and cudf https://github.com/rapidsai/cudf/issues/17649
        request.applymarker(pytest.mark.xfail)
    df = nw.from_native(constructor({"a": [1, 1, 2], "b": [4, 5, 6]}))
    result = df.group_by("a").agg(c=nw.col("b").mean(), d=nw.col("b").mean()).sort("a")
    expected = {"a": [1, 2], "c": [4.5, 6], "d": [4.5, 6]}
    assert_equal_data(result, expected)


def test_all_kind_of_aggs(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("dask",)):
        # bugged in dask https://github.com/dask/dask/issues/11612
        # and cudf https://github.com/rapidsai/cudf/issues/17649
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor) and PANDAS_VERSION < (1, 4):
        pytest.skip(
            "Pandas < 1.4.0 does not support multiple aggregations with the same column"
        )
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    df = nw.from_native(constructor({"a": [1, 1, 1, 2, 2, 2], "b": [4, 5, 6, 0, 5, 5]}))
    result = (
        df.group_by("a")
        .agg(
            c=nw.col("b").mean(),
            d=nw.col("b").mean(),
            e=nw.col("b").std(ddof=1),
            f=nw.col("b").std(ddof=2),
            g=nw.col("b").var(ddof=2),
            h=nw.col("b").var(ddof=2),
            i=nw.col("b").n_unique(),
        )
        .sort("a")
    )

    variance_num = sum((v - 10 / 3) ** 2 for v in [0, 5, 5])
    expected = {
        "a": [1, 2],
        "c": [5, 10 / 3],
        "d": [5, 10 / 3],
        "e": [1, (variance_num / (3 - 1)) ** 0.5],
        "f": [2**0.5, (variance_num) ** 0.5],  # denominator is 1 (=3-2)
        "g": [2.0, variance_num],  # denominator is 1 (=3-2)
        "h": [2.0, variance_num],  # denominator is 1 (=3-2)
        "i": [3, 2],
    }
    assert_equal_data(result, expected)


def test_pandas_group_by_index_and_column_overlap() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df = pd.DataFrame(
        {"a": [1, 1, 2], "b": [4, 5, 6]}, index=pd.Index([0, 1, 2], name="a")
    )
    result = nw.from_native(df, eager_only=True).group_by("a").agg(nw.col("b").mean())
    expected = {"a": [1, 2], "b": [4.5, 6.0]}
    assert_equal_data(result, expected)

    key, result = next(iter(nw.from_native(df, eager_only=True).group_by("a")))
    assert key == (1,)
    expected_native = pd.DataFrame({"a": [1, 1], "b": [4, 5]}, index=pd.Index([0, 1]))
    pd.testing.assert_frame_equal(result.to_native(), expected_native)


def test_fancy_functions(constructor: Constructor) -> None:
    df = nw.from_native(constructor({"a": [1, 1, 2], "b": [4, 5, 6]}))
    result = df.group_by("a").agg(nw.all().std(ddof=0)).sort("a")
    expected = {"a": [1, 2], "b": [0.5, 0.0]}
    assert_equal_data(result, expected)
    result = df.group_by("a").agg(nw.selectors.numeric().std(ddof=0)).sort("a")
    assert_equal_data(result, expected)
    result = df.group_by("a").agg(nw.selectors.matches("b").std(ddof=0)).sort("a")
    assert_equal_data(result, expected)
    result = (
        df.group_by("a").agg(nw.selectors.matches("b").std(ddof=0).alias("c")).sort("a")
    )
    expected = {"a": [1, 2], "c": [0.5, 0.0]}
    assert_equal_data(result, expected)
    result = (
        df.group_by("a")
        .agg(nw.selectors.matches("b").std(ddof=0).name.map(lambda _x: "c"))
        .sort("a")
    )
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("keys", "aggs", "expected", "sort_by"),
    [
        (
            [nw.col("a").abs(), nw.col("a").abs().alias("a_with_alias")],
            [nw.col("x").sum()],
            {"a": [1, 2], "a_with_alias": [1, 2], "x": [5, 5]},
            ["a"],
        ),
        (
            [nw.col("a").alias("x")],
            [nw.col("x").mean().alias("y")],
            {"x": [-1, 1, 2], "y": [4.0, 0.5, 2.5]},
            ["x"],
        ),
        (
            [nw.col("a")],
            [nw.col("a").count().alias("foo-bar"), nw.all().sum()],
            {"a": [-1, 1, 2], "foo-bar": [1, 2, 2], "x": [4, 1, 5], "y": [1.5, 0, 0]},
            ["a"],
        ),
        (
            [nw.col("a", "y").abs()],
            [nw.col("x").sum()],
            {"a": [1, 1, 2], "y": [0.5, 1.5, 1], "x": [1, 4, 5]},
            ["a", "y"],
        ),
        (
            [nw.col("a").abs().alias("y")],
            [nw.all().sum().name.suffix("c")],
            {"y": [1, 2], "ac": [1, 4], "xc": [5, 5]},
            ["y"],
        ),
        (
            [nw.selectors.by_dtype(nw.Float64()).abs()],
            [nw.selectors.numeric().sum()],
            {"y": [0.5, 1.0, 1.5], "a": [2, 4, -1], "x": [1, 5, 4]},
            ["y"],
        ),
    ],
)
def test_group_by_expr(
    constructor: Constructor,
    keys: list[nw.Expr],
    aggs: list[nw.Expr],
    expected: dict[str, list[Any]],
    sort_by: list[str],
) -> None:
    data = {"a": [1, 1, 2, 2, -1], "x": [0, 1, 2, 3, 4], "y": [0.5, -0.5, 1.0, -1.0, 1.5]}
    df = nw.from_native(constructor(data))
    result = df.group_by(*keys).agg(*aggs).sort(*sort_by)
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    "keys",
    [
        [nw.col("a").drop_nulls()],  # Transform and Filtration
        [nw.col("a").alias("foo"), nw.col("a").drop_nulls()],  # Transform and Filtration
        [nw.col("a").alias("foo"), nw.col("a").max()],  # Transform and Aggregation
        [nw.lit(42)],  # Literal
        [nw.lit(42).abs()],  # Literal
    ],
)
def test_group_by_raise_if_not_preserves_length(
    constructor: Constructor, keys: list[nw.Expr]
) -> None:
    data = {"a": [1, 2, 2, None], "b": [0, 1, 2, 3], "x": [1, 2, 3, 4]}
    df = nw.from_native(constructor(data))
    with pytest.raises((InvalidOperationError, NotImplementedError)):
        df.group_by(keys).agg(nw.col("x").max())


def test_group_by_window(constructor: Constructor) -> None:
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 2, 2, None], "b": [1, 1, 2, 2], "x": [1, 2, 3, 4]}
    df = nw.from_native(constructor(data))
    result = (
        df.group_by(nw.col("a").mean().over("b"))
        .agg(nw.col("x").max())
        .sort("a", nulls_last=True)
    )
    expected = {"a": [1.5, 2.0], "x": [2, 4]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    "keys", [[nw.col("a").abs()], ["a", nw.col("a").abs().alias("a_test")]]
)
def test_group_by_raise_drop_null_keys_with_exprs(
    constructor: Constructor, keys: list[nw.Expr | str]
) -> None:
    data = {"a": [1, 1, 2, 2, -1], "x": [0, 1, 2, 3, 4], "y": [0.5, -0.5, 1.0, -1.0, 1.5]}
    df = nw.from_native(constructor(data))
    with pytest.raises(
        NotImplementedError, match="drop_null_keys cannot be True when keys contains Expr"
    ):
        df.group_by(*keys, drop_null_keys=True)  # type: ignore[call-overload]


def test_group_by_selector(constructor: Constructor) -> None:
    data = {
        "a": [1, 1, 1],
        "b": [4, 4, 6],
        "c": ["foo", "foo", "bar"],
        "x": [7.5, 8.5, 9.0],
    }
    result = (
        nw.from_native(constructor(data))
        .group_by(nw.selectors.by_dtype(nw.Int64), "c")
        .agg(nw.col("x").mean())
        .sort("a", "b")
    )
    expected = {"a": [1, 1], "b": [4, 6], "c": ["foo", "bar"], "x": [8.0, 9.0]}
    assert_equal_data(result, expected)


def test_renaming_edge_case(constructor: Constructor) -> None:
    data = {"a": [0, 0, 0], "_a_tmp": [1, 2, 3], "b": [4, 5, 6]}
    result = nw.from_native(constructor(data)).group_by(nw.col("a")).agg(nw.all().min())
    expected = {"a": [0], "_a_tmp": [1], "b": [4]}
    assert_equal_data(result, expected)


def test_group_by_len_1_column(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    """Based on a failure from marimo.

    - https://github.com/marimo-team/marimo/blob/036fd3ff89ef3a0e598bebb166637028024f98bc/tests/_plugins/ui/_impl/tables/test_narwhals.py#L1098-L1108
    - https://github.com/marimo-team/marimo/blob/036fd3ff89ef3a0e598bebb166637028024f98bc/marimo/_plugins/ui/_impl/tables/narwhals_table.py#L163-L188
    """
    if any(x in str(constructor) for x in ("dask",)):
        # `dask`
        #     ValueError: conflicting aggregation functions: [('size', 'a'), ('size', 'a')]
        request.applymarker(pytest.mark.xfail)
    data = {"a": [1, 2, 1, 2, 3, 4]}
    expected = {"a": [1, 2, 3, 4], "len": [2, 2, 1, 1], "len_a": [2, 2, 1, 1]}
    result = (
        nw.from_native(constructor(data))
        .group_by("a")
        .agg(nw.len(), nw.len().alias("len_a"))
        .sort("a")
    )
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("low", "high"),
    [
        ("A", "B"),
        (1.5, 5.2),
        (dt.datetime(2000, 1, 1), dt.datetime(2002, 1, 1)),
        (dt.date(2000, 1, 1), dt.date(2002, 1, 1)),
        (dt.time(5, 0, 0), dt.time(14, 0, 0)),
        (dt.timedelta(32), dt.timedelta(800)),
        (False, True),
        (b"a", b"z"),
        (Decimal("43.954"), Decimal("264.124")),
    ],
    ids=[
        "str",
        "float",
        "datetime",
        "date",
        "time",
        "timedelta",
        "bool",
        "bytes",
        "Decimal",
    ],
)
def test_group_by_no_preserve_dtype(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    low: NonNestedLiteral,
    high: NonNestedLiteral,
) -> None:
    """Minimal repro for [`px.sunburst` failure].

    The issue appeared for `n_unique`, but applies for any [aggregation that requires a function].

    [`px.sunburst` failure]: https://github.com/narwhals-dev/narwhals/pull/2680#discussion_r2151972940
    [aggregation that requires a function]: https://github.com/pandas-dev/pandas/issues/57317
    """
    if (
        "polars" in str(constructor_eager)
        and isinstance(low, Decimal)
        and POLARS_VERSION < (1, 21, 0)
    ):
        pytest.skip("Decimal support in group_by for polars didn't stabilize until 1.0.0")
    if any(x == request.node.callspec.id for x in ("cudf-time", "cudf-bytes")):
        request.applymarker(pytest.mark.xfail)

    data = {
        "col_a": ["A", "B", None, "A", "A", "B", None],
        "col_b": [low, low, high, high, None, None, None],
    }
    expected = {"col_a": [None, "A", "B"], "n_unique": [2, 3, 2]}
    frame = nw.from_native(constructor_eager(data))
    result = (
        frame.group_by("col_a").agg(n_unique=nw.col("col_b").n_unique()).sort("col_a")
    )
    actual_dtype = result.schema["n_unique"]
    assert actual_dtype.is_integer()
    assert_equal_data(result, expected)


def test_top_level_len(constructor: Constructor) -> None:
    # https://github.com/holoviz/holoviews/pull/6567#issuecomment-3178743331
    df = nw.from_native(
        constructor({"gender": ["m", "f", "f"], "weight": [4, 5, 6], "age": [None, 8, 9]})
    )
    result = df.group_by(["gender"]).agg(nw.all().len()).sort("gender")
    expected = {"gender": ["f", "m"], "weight": [2, 1], "age": [2, 1]}
    assert_equal_data(result, expected)
    result = (
        df.group_by("gender")
        .agg(nw.col("weight").len(), nw.col("age").len())
        .sort("gender")
    )
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("keys", "aggs", "expected", "pre_sort"),
    [
        (["a"], ["b"], {"a": [1, 2, 3, 4], "b": [1, 2, 4, 6]}, None),
        (["a"], ["b"], {"a": [1, 2, 3, 4], "b": [1, 3, 5, 6]}, {"descending": True}),
        (["a"], ["c"], {"a": [1, 2, 3, 4], "c": [None, "A", None, "B"]}, None),
        (
            ["a"],
            ["c"],
            {"a": [1, 2, 3, 4], "c": [None, "A", "B", "B"]},
            {"nulls_last": True},
        ),
    ],
    ids=["no-sort", "sort-descending", "NA-order-nulls-first", "NA-order-nulls-last"],
)
def test_group_by_agg_first(
    constructor_eager: ConstructorEager,
    keys: Sequence[str],
    aggs: Sequence[str],
    expected: Mapping[str, Any],
    pre_sort: Mapping[str, Any] | None,
    request: pytest.FixtureRequest,
) -> None:
    request.applymarker(
        pytest.mark.xfail(
            "pyarrow_table" in str(constructor_eager) and (PYARROW_VERSION < (14, 0)),
            reason="https://github.com/apache/arrow/issues/36709",
            raises=NotImplementedError,
        )
    )
    data = {
        "a": [1, 2, 2, 3, 3, 4],
        "b": [1, 2, 3, 4, 5, 6],
        "c": [None, "A", "A", None, "B", "B"],
    }
    df = nw.from_native(constructor_eager(data))
    if pre_sort:
        df = df.sort(aggs, **pre_sort)
    result = df.group_by(keys).agg(nw.col(aggs).first()).sort(keys)
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("keys", "aggs", "expected", "pre_sort"),
    [
        (["a"], ["b"], {"a": [1, 2, 3, 4], "b": [1, 3, 5, 6]}, None),
        (["a"], ["b"], {"a": [1, 2, 3, 4], "b": [1, 2, 4, 6]}, {"descending": True}),
        (["a"], ["c"], {"a": [1, 2, 3, 4], "c": [None, "A", "B", "B"]}, None),
        (
            ["a"],
            ["c"],
            {"a": [1, 2, 3, 4], "c": [None, "A", None, "B"]},
            {"nulls_last": True},
        ),
    ],
    ids=["no-sort", "sort-descending", "NA-order-nulls-first", "NA-order-nulls-last"],
)
def test_group_by_agg_last(
    constructor_eager: ConstructorEager,
    keys: Sequence[str],
    aggs: Sequence[str],
    expected: Mapping[str, Any],
    pre_sort: Mapping[str, Any] | None,
    request: pytest.FixtureRequest,
) -> None:
    request.applymarker(
        pytest.mark.xfail(
            "pyarrow_table" in str(constructor_eager) and (PYARROW_VERSION < (14, 0)),
            reason="https://github.com/apache/arrow/issues/36709",
            raises=NotImplementedError,
        )
    )
    data = {
        "a": [1, 2, 2, 3, 3, 4],
        "b": [1, 2, 3, 4, 5, 6],
        "c": [None, "A", "A", None, "B", "B"],
    }
    df = nw.from_native(constructor_eager(data))
    if pre_sort:
        df = df.sort(aggs, **pre_sort)
    result = df.group_by(keys).agg(nw.col(aggs).last()).sort(keys)
    assert_equal_data(result, expected)


def test_multi_column_expansion(constructor: Constructor) -> None:
    if "polars" in str(constructor) and POLARS_VERSION < (1, 32):
        pytest.skip(reason="https://github.com/pola-rs/polars/issues/21773")
    if "modin" in str(constructor):
        pytest.skip(reason="Internal error")
    df = nw.from_native(constructor({"a": [1, 1, 2], "b": [4, 5, 6]}))
    result = (
        df.group_by("a")
        .agg(nw.all().sum().name.suffix("_aggregated"))
        .sort("a", descending=True)
    )
    expected = {"a": [2, 1], "b_aggregated": [6, 9]}
    assert_equal_data(result, expected)
    result = (
        df.group_by("a")
        .agg(nw.col("a", "b").sum().name.suffix("_aggregated"))
        .sort("a", descending=True)
    )
    expected = {"a": [2, 1], "a_aggregated": [2, 2], "b_aggregated": [6, 9]}
    assert_equal_data(result, expected)
    result = (
        df.group_by("a")
        .agg(nw.nth(0, 1).sum().name.suffix("_aggregated"))
        .sort("a", descending=True)
    )
    assert_equal_data(result, expected)



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import ConstructorEager, assert_equal_data


@pytest.mark.parametrize("n", [2, -1])
def test_tail_series(constructor_eager: ConstructorEager, n: int) -> None:
    df = nw.from_native(constructor_eager({"a": [1, 2, 3]}), eager_only=True)
    result = df.select(df["a"].tail(n))
    expected = {"a": [2, 3]}
    assert_equal_data(result, expected)



"""Generate database and answers via [DuckDB TPC-H Extension].

[DuckDB TPC-H Extension]: https://duckdb.org/docs/stable/core_extensions/tpch
"""

from __future__ import annotations

import argparse
import dataclasses
import logging
import os
import sys
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

from tpch import constants
from tpch.classes import TableLogger

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping

    import polars as pl
    import pytest
    from duckdb import DuckDBPyConnection as Con, DuckDBPyRelation as Rel
    from typing_extensions import LiteralString

    from tpch.typing_ import Artifact, QueryID, ScaleFactor


logger = logging.getLogger(constants.LOGGER_NAME)

TABLE_SCALE_FACTOR = """\
┌───────┬────────────┬─────────────┐
│ sf    ┆ Disk       ┆ Memory (db) │
╞═══════╪════════════╪═════════════╡
│ 0.014 ┆    3.25 mb ┆    8.79 mb  │
│ 0.052 ┆   12.01 mb ┆   32.49 mb  │
│ 0.1   ┆   23.15 mb ┆   62.62 mb  │
│ 0.25  ┆   58.90 mb ┆  159.32 mb  │
│ 0.51  ┆  124.40 mb ┆  336.50 mb  │
│ 1.0   ┆  247.66 mb ┆  669.92 mb  │
│ 10.0  ┆    2.59 gb ┆    7.00 gb  │
│ 30.0  ┆    7.76 gb ┆   21.00 gb  │
└───────┴────────────┴─────────────┘
"""

FIX_ANSWERS: Mapping[QueryID, Callable[[pl.LazyFrame], pl.LazyFrame]] = {
    "q18": lambda df: df.rename({"sum(l_quantity)": "sum"}).cast({"sum": int}),
    "q22": lambda df: df.cast({"cntrycode": int}),
}
"""
DuckDB being weird, this is [correct] but [not this one].

[correct]: https://github.com/duckdb/duckdb/blob/47c227d7d8662586b0307d123c03b25c0db3d515/extension/tpch/dbgen/answers/sf0.01/q18.csv#L1
[not this one]: https://github.com/duckdb/duckdb/blob/47c227d7d8662586b0307d123c03b25c0db3d515/extension/tpch/dbgen/answers/sf100/q18.csv#L1
"""


def read_fmt_schema(fp: Path) -> str:
    import polars as pl

    schema = pl.read_parquet_schema(fp).items()
    return f"- {fp.name}\n" + "\n".join(f"  - {k:<20}: {v}" for k, v in schema)


@cache
def cast_map() -> dict[Any, Any]:
    import polars as pl
    import polars.selectors as cs

    return {cs.decimal(): float, cs.date(): pl.Datetime("ns")}


@dataclasses.dataclass(**({"kw_only": True} if sys.version_info >= (3, 10) else {}))
class TPCHGen:
    scale_factor: ScaleFactor
    refresh: bool = False
    debug: bool = False
    _con: Con = dataclasses.field(init=False, repr=False)

    @staticmethod
    def from_pytest(config: pytest.Config, /) -> TPCHGen:
        return TPCHGen(scale_factor=config.getoption("--scale-factor"))

    @staticmethod
    def from_argparse(parser: argparse.ArgumentParser, /) -> TPCHGen:
        return parser.parse_args(namespace=TPCHGen.__new__(TPCHGen))

    @property
    def scale_factor_dir(self) -> Path:
        return constants._scale_factor_dir(self.scale_factor)

    @property
    def _database(self) -> Path | Literal[":memory:"]:
        return constants.DB_PATH if self.scale_factor == "30.0" else ":memory:"

    def glob(self, artifact: Artifact, /) -> Iterator[Path]:
        return self.scale_factor_dir.glob(constants.GLOBS[artifact])

    def has_data(self) -> bool:
        both = next(self.glob("answers"), None) and next(self.glob("database"), None)
        return bool(both)

    def run(self) -> None:
        _configure_logger(debug=self.debug)
        if self.refresh:
            logger.info("Refreshing data for scale_factor=%s", self.scale_factor)
        elif self.has_data():
            print(  # noqa: T201
                f"Reusing cached TPC-H data\n"
                f"To regenerate this dataset, use `--scale_factor={self.scale_factor} --refresh`"
            )
            self.show_schemas("database").show_schemas("answers")
            return
        self.connect().load_extension().generate_database().write_database().write_answers().disconnect()
        n_bytes = sum(e.stat().st_size for e in os.scandir(self.scale_factor_dir))
        total = TableLogger.format_size(n_bytes)
        logger.info("Finished with total file size: %s", total.strip())

    def connect(self) -> TPCHGen:
        import duckdb

        database = self._database
        name = database.as_posix() if isinstance(database, Path) else database
        logger.info("Connecting to DuckDB database: %s", name)
        self._con = duckdb.connect(database)
        return self

    def disconnect(self) -> TPCHGen:
        self._con.close()
        if isinstance(self._database, Path):
            logger.info("Dropping DuckDB database: %s", self._database.as_posix())
            self._database.unlink(missing_ok=True)
        return self

    def sql(self, query: LiteralString, **params: LiteralString | int) -> Rel:
        return self._con.sql(query, params=params or None)

    def load_extension(self) -> TPCHGen:
        logger.info("Installing DuckDB TPC-H Extension")
        self._con.install_extension("tpch")
        self._con.load_extension("tpch")
        return self

    def _generate_database_batched(self, batches: int) -> TPCHGen:
        logger.info("Whelp, this may take a while...")
        logger.info("Generating in %s batches", batches)
        query = "CALL dbgen(sf=$sf, children=$children, step=$step)"
        for batch in range(batches):
            self.sql(query, sf=self.scale_factor, children=batches, step=batch)
            logger.info("Generated (%s/%s)", batch + 1, batches)
        return self

    def generate_database(self) -> TPCHGen:
        sf = self.scale_factor
        logger.info("Generating data for scale_factor=%s", sf)
        if sf in {"10.0", "30.0"}:
            # NOTE: See description https://github.com/narwhals-dev/narwhals/pull/3429#issue-3879052445
            self._generate_database_batched(12 if sf == "10.0" else 4)
        else:
            self.sql("CALL dbgen(sf=$sf)", sf=sf)
        logger.info("Finished generating data.")
        if logger.isEnabledFor(logging.DEBUG):
            query = """SELECT "table": name, "schema": MAP(column_names, column_types) FROM (SHOW ALL TABLES)"""
            logger.debug("DuckDB schemas (database):\n%s", str(self.sql(query))[:-1])
        return self

    def write_database(self) -> TPCHGen:
        logger.info("Writing data to: %s", self.scale_factor_dir.as_posix())
        with TableLogger.database() as tbl_logger:
            for t in constants.DATABASE_TABLE_NAMES:
                path = self.scale_factor_dir / f"{t}.parquet"
                to_polars(self.sql(f"FROM {t}")).sink_parquet(path)
                tbl_logger.log_row(path)
        return self.show_schemas("database")

    def write_answers(self) -> TPCHGen:
        logger.info("Executing TPC-H queries for answers")
        with TableLogger.answers() as tbl_logger:
            for query_id in constants.QUERY_IDS:
                lf = to_polars(self.sql(f"PRAGMA tpch({query_id.removeprefix('q')})"))
                if fix := FIX_ANSWERS.get(query_id):
                    lf = fix(lf)
                path = self.scale_factor_dir / f"result_{query_id}.parquet"
                lf.sink_parquet(path)
                tbl_logger.log_row(path)
        return self.show_schemas("answers")

    def show_schemas(self, artifact: Artifact, /) -> TPCHGen:
        if logger.isEnabledFor(logging.DEBUG):
            if paths := sorted(self.glob(artifact)):
                msg = "\n".join(read_fmt_schema(fp) for fp in paths)
                logger.debug("Parquet schemas (%s):\n%s", artifact, msg)
            else:
                msg = f"Found no matching paths for {artifact!r} in {self.scale_factor_dir.as_posix()}"
                raise NotImplementedError(msg)
        return self


def to_polars(rel: Rel) -> pl.LazyFrame:
    return rel.pl(lazy=True).cast(cast_map())


def _configure_logger(
    *,
    debug: bool,
    fmt: str = "%(asctime)s.%(msecs)03d [%(levelname)s] %(message)s",
    datefmt: str = "%H:%M:%S",
) -> None:
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    output = logging.StreamHandler()
    output.setFormatter(logging.Formatter(fmt, datefmt))
    logger.addHandler(output)


class HelpFormatter(
    argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
):
    def start_section(self, heading: str | None) -> None:
        super().start_section(_title.capitalize() if (_title := heading) else heading)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=HelpFormatter,
        description="Generate the data required to run TPC-H queries.\n\nUsage: %(prog)s [OPTIONS]",
        usage=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-sf",
        "--scale-factor",
        default=constants.SCALE_FACTOR_DEFAULT,
        metavar="",
        help=f"Scale the database by this factor (default: %(default)s)\n{TABLE_SCALE_FACTOR}",
        choices=constants.SCALE_FACTORS,
    )
    parser.add_argument(
        "--debug", action="store_true", help="Enable more detailed logging"
    )
    parser.add_argument(
        "--refresh",
        action="store_true",
        help="Re-run data generation, regardless of whether `--scale-factor` is already on disk",
    )
    TPCHGen.from_argparse(parser).run()



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(part_ds: FrameT, partsupp_ds: FrameT, supplier_ds: FrameT) -> FrameT:
    var1 = "Brand#45"

    supplier = supplier_ds.filter(
        nw.col("s_comment").str.contains(".*Customer.*Complaints.*")
    ).select(nw.col("s_suppkey"), nw.col("s_suppkey").alias("ps_suppkey"))

    return (
        part_ds.join(partsupp_ds, left_on="p_partkey", right_on="ps_partkey")
        .filter(nw.col("p_brand") != var1)
        .filter(~nw.col("p_type").str.contains("MEDIUM POLISHED*"))
        .filter(nw.col("p_size").is_in([49, 14, 23, 45, 19, 3, 36, 9]))
        .join(supplier, left_on="ps_suppkey", right_on="s_suppkey", how="left")  # pyright: ignore[reportArgumentType]
        .filter(nw.col("ps_suppkey_right").is_null())
        .group_by("p_brand", "p_type", "p_size")
        .agg(nw.col("ps_suppkey").n_unique().alias("supplier_cnt"))
        .sort(
            by=["supplier_cnt", "p_brand", "p_type", "p_size"],
            descending=[True, False, False, False],
        )
    )



from __future__ import annotations

from narwhals.stable.v1._dtypes import (
    Array,
    Binary,
    Boolean,
    Categorical,
    Date,
    Datetime,
    Decimal,
    DType,
    Duration,
    Enum,
    Field,
    Float32,
    Float64,
    FloatType,
    Int8,
    Int16,
    Int32,
    Int64,
    Int128,
    IntegerType,
    List,
    NestedType,
    NumericType,
    Object,
    SignedIntegerType,
    String,
    Struct,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    Unknown,
    UnsignedIntegerType,
)

__all__ = [
    "Array",
    "Binary",
    "Boolean",
    "Categorical",
    "DType",
    "Date",
    "Datetime",
    "Decimal",
    "Duration",
    "Enum",
    "Field",
    "Float32",
    "Float64",
    "FloatType",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Int128",
    "IntegerType",
    "List",
    "NestedType",
    "NumericType",
    "Object",
    "SignedIntegerType",
    "String",
    "Struct",
    "Time",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt128",
    "Unknown",
    "UnsignedIntegerType",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals.stable.v1.dependencies import is_into_dataframe

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typing_extensions import Self

DATA: dict[str, Any] = {"a": [1, 2, 3], "b": [4, 5, 6]}


class DictDataFrame:
    def __init__(self, data: Mapping[str, Any]) -> None:
        self._data = data

    def __len__(self) -> int:  # pragma: no cover
        return len(next(iter(self._data.values())))

    def __narwhals_dataframe__(self) -> Self:  # pragma: no cover
        return self


def test_is_into_dataframe_pyarrow() -> None:
    pytest.importorskip("pyarrow")
    import pyarrow as pa

    assert is_into_dataframe(pa.table(DATA))


def test_is_into_dataframe_polars() -> None:
    pytest.importorskip("polars")
    import polars as pl

    assert is_into_dataframe(pl.DataFrame(DATA))


def test_is_into_dataframe_pandas() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    assert is_into_dataframe(pd.DataFrame(DATA))
    assert is_into_dataframe(nw.from_native(pd.DataFrame(DATA)))


def test_is_into_dataframe_other() -> None:
    pytest.importorskip("numpy")
    import numpy as np

    assert is_into_dataframe(DictDataFrame(DATA))
    assert not is_into_dataframe(np.array([[1, 4], [2, 5], [3, 6]]))
    assert not is_into_dataframe(DATA)



from __future__ import annotations

from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, get_args

from tpch.typing_ import Artifact, DBTableName, QueryID, ScaleFactor

if TYPE_CHECKING:
    from collections.abc import Mapping

DATA_DIR = Path(__file__).parent / "data"
DB_PATH = DATA_DIR / "narwhals.duckdb"


@cache
def _scale_factor_dir(scale_factor: ScaleFactor) -> Path:
    """Get the data directory for a specific scale factor."""
    sf_dir = DATA_DIR / f"sf{scale_factor}"
    sf_dir.mkdir(parents=True, exist_ok=True)
    return sf_dir


SCALE_FACTORS: tuple[ScaleFactor, ...] = get_args(ScaleFactor)
SCALE_FACTOR_DEFAULT: ScaleFactor = "0.1"
DATABASE_TABLE_NAMES: tuple[DBTableName, ...] = get_args(DBTableName)
QUERY_IDS: tuple[QueryID, ...] = get_args(QueryID)
GLOBS: Mapping[Artifact, str] = {
    "database": r"*[!0-9].parquet",
    "answers": r"result_q[0-9]*.parquet",
}
LOGGER_NAME = "narwhals.tpch"
"""Per-[Logging Cookbook], pass this to `logging.getLogger(...)`.

[Logging Cookbook]: https://docs.python.org/3/howto/logging-cookbook.html#using-loggers-as-attributes-in-a-class-or-passing-them-as-parameters
"""
QUERIES_PACKAGE = "tpch.queries"
